/*
 ķ
   Program..: _DCXBROW.PRG                                 
   Author...:  Roger J. Donnay                             
   Notice...: (c) DONNAY Software Designs 1987-2002        
   Date.....: Jan 16, 2002                                 
   Notes....: eXPress++ Browse Classes                     
 Ľ
*/

#INCLUDE "appevent.ch"
#INCLUDE "dcdialog.ch"
#INCLUDE "dll.ch"
#INCLUDE "dcbitmap.CH"
#INCLUDE "dcads.CH"

#DEFINE QUERY_SAY          1
#DEFINE QUERY_FIELD        2
#DEFINE QUERY_PICTURE      3
#DEFINE QUERY_INDEX        4
#DEFINE QUERY_ROW          5
#DEFINE QUERY_COL          6
#DEFINE QUERY_GET          7
#DEFINE QUERY_VAR          8

// Drawing modes used with XBPs
#ifndef XBP_DRAW_NORMAL
#define XBP_DRAW_NORMAL                0
#define XBP_DRAW_OWNER                 1
#define XBP_DRAW_OWNERADVANCED         2
#endif

CLASS DC_XbpBrowse FROM DC_XbpBrowseSuper, DC_XbaseParts

EXPORTED:
  VAR oHasFocus
  VAR dataLink
  VAR color
  VAR varBlock
  VAR captionBlock
  VAR picture
  VAR optionArray
  VAR arrayElement
  VAR dataSource
  VAR dataSourceType                      // J. Duijf 26-11-2002
  VAR thumbRecords
  VAR headerLines
  VAR headerDelim
  VAR footerLines
  VAR footerDelim
  VAR headerPres
  VAR footerPres
  VAR browsePres                          // J. Duijf 10-03-2011
  VAR fit
  VAR fitMax
  VAR optimize
  VAR outputBlock
  VAR inputBlock
  VAR mark
  VAR markColorBlock
  VAR rowColorBlock
  VAR isScope
  VAR font
  VAR pointerBlock
  VAR autoLock
  VAR freezeLeft
  VAR freezeRight
  VAR columns
  VAR sortColorSelectedFG
  VAR sortColorSelectedBG
  VAR sortColorUnSelectedFG
  VAR sortColorUnSelectedBG
  VAR sortColorNonselectableFG
  VAR sortColorNonselectableBG
  VAR descending
  VAR descendToggle
  VAR sortImageAscend
  VAR sortImageDescend
  VAR sortColumn
  VAR refreshInterval
  VAR refreshWhenBlock
  VAR refreshBlock
  VAR timer
  VAR rightButtonSelect
  VAR cellEditMode
  VAR cellEditor
  VAR saveRecord
  VAR hasHeader
  VAR hasFooter
  VAR navBlocks
  VAR dataEnabled
  VAR rightButtonDown
  VAR defaultColumnsGetList
  VAR configMessageBlock
  VAR filter
  VAR noRestore
  VAR restoreConfig
  VAR restoreProc
  VAR skipUnique
  VAR lBrow
  VAR nBrow
  VAR resizeColumns
  VAR tagEnable
  VAR tagElement
  VAR tagColorFG
  VAR tagColorBG
  VAR tagMode
  VAR lastTagRec
  VAR lastTagPos
  VAR config
  VAR scrollBarTipBlock
  VAR isItemMarkedCodeBlock
  VAR ownerDrawBlock
  VAR isLeftMouseButton
  VAR isTrapErrors
  VAR evenRow
  VAR CellEditKeyBlock                                                        // J. Duijf 30-03-2013
  VAR isCellEditing
  VAR scrollBarHeight
  VAR navigateBlock
  VAR recordObject

  METHOD Init, create, Navigate, FieldBlock, CellEditSetup, ;
         Save, Restore, MoveColumn, Refresh, WhenHideEval, Destroy, ;
         handleEvent, setTimerRefresh, right, itemRbDown, CellEdit, ;
         CellEditMode, RefreshAll, SortClear, HFLbDown, DataDisplay, ;
         Optimize, RestoreDefaults, FitColumns, ForceStable, ;
         RefreshView, RowColor, ResizeColumns, Tag, Resize, SetSize

  METHOD XbpBrowsehandleEvent IS handleEvent IN DC_XbpBrowseSuper
  METHOD ZebraColor                                                           // J. Duijf 21-04-2010
  METHOD CustomDrawCell

ENDCLASS

* ---------------------

METHOD DC_XbpBrowse:init( oParent, oOwner, aPos, aSize, aPres, lVisible, oGetList )

LOCAL xOptions, bBlock, aArray, bColorBlock, aColor, nPointer, ;
      aGetList, aGetListItem, xData, aSort, xFont, lThumbLock, ;
      nFound, aConfig, cID, nColumns
LOCAL aEvenColor, aOddColor, bTmpColorBlock

::DC_XbaseParts:init( oGetList,,self )

aGetList := ::getList:getListArray
aGetListItem := ::getList:getListArray[::getListPointer]

IF Valtype(aGetListItem[oGETLIST_CONFIG]) == 'O'
  ::config := aGetListItem[oGETLIST_CONFIG]
ELSE
  ::config := DC_XbpBrowseDefault()
ENDIF

//
// J. Duijf 11-03-2011
// First get the default presentation parameters (As in DC_BrowPres())
// Then merge the optional presentation parameters with the defaults.
//
::browsePres      := Aclone(::config:presParam)                               // J. Duijf 10-03-2011
DC_MergePresentation(::browsePres,aPres)

::browsePres      := ApplyScaleFactor(::browsePres,oGetList)
//
// J. Duijf, 12-04-2011 Moved here to update presentation parameters
//
xFont             := aGetListItem[cGETLIST_FONT]
IF Empty(xFont)
  DEFAULT xFont   := ::config:font
ENDIF

IF !Empty(xFont)
  IF Valtype(xFont) == 'B'
    xFont := Eval(xFont)
  ENDIF
   DC_MergePresentation(::browsePres,{{XBP_PP_COL_DA_COMPOUNDNAME,xFont}})
   DC_MergePresentation(::browsePres,{{XBP_PP_COL_HA_COMPOUNDNAME,xFont}})
   DC_MergePresentation(::browsePres,{{XBP_PP_COL_FA_COMPOUNDNAME,xFont}})
//  ::setFontCompoundName(Alltrim(xFont))
ENDIF


bBlock := DC_GetBlock(aGetListItem[bGETLIST_REFVAR])
IF Valtype(bBlock) = 'B'
  xData := Eval( bBlock )
  IF !DC_ReadDataCheck( xData, 'ACONI', aGetList, ::getListPointer , .t. )
    RETURN nil
  ENDIF
ELSEIF Valtype(aGetListItem[cGETLIST_VARNAME]) = 'C'
  xData := aGetListItem[cGETLIST_VARNAME]
  IF !DC_ReadDataCheck( xData, 'C', aGetList, ::getListPointer, .t. )
    RETURN nil
  ENDIF
ENDIF

IF Valtype(xData) # 'A' .AND. Empty(xData)
  xData := Alias()
  IF Empty(xData)
    DC_WinAlert('DC_ReadGui() error! Browse has no data source')
    RETURN nil
  ENDIF
ENDIF

xOptions := aGetListItem[xGETLIST_OPTIONS]
ASize(xOptions,53)
aArray := xOptions

::XbpBrowse:init( oParent, oOwner, aPos, aSize, ::browsePres, lVisible )

IF Valtype(aArray[51]) == 'N'
  ::nScrollHeight := aArray[51]
  ::oHScroll:setSize({::oHScroll:currentSize()[1],::nScrollHeight})
  ::oVScroll:setSize({::nScrollHeight,::oVScroll:currentSize()[2]})
ENDIF

::scrollBarHeight := ::nScrollHeight // protected var in XbpBrowse() class.

IF IsMemberVar(self,'useVisualStyle')
  ::useVisualStyle := ::config:useVisualStyleBrowse
ENDIF

::tabstop := .f.
::dataEnabled := .t.
::noRestore := .f.
::skipUnique := .f.
::nBrow := 1
::lBrow := .f.
::isItemMarkedCodeBlock := .f.
::isTrapErrors := DC_BrowseTrapErrors()
::evenRow := .f.
::isLeftMouseButton := .f.
::configMessageBlock := DC_BrowseConfigMessageBlock()
::isCellEditing := .f.

IF Valtype(aGetListItem[xGETLIST_OPTIONS7]) = 'N'
  ::cursorMode := aGetListItem[xGETLIST_OPTIONS7]
ELSE
  ::cursorMode := XBPBRW_CURSOR_CELL
ENDIF

IF Valtype(xData) = 'A'
  aGetListItem[nGETLIST_SUBTYPE] := BROWSE_ARRAY
ELSEIF Valtype(xData) = 'C'
  aGetListItem[nGETLIST_SUBTYPE] := BROWSE_DATABASE
ELSEIF Valtype(xData) = 'O'                                                   // J. Duijf 27-11-2002
   IF xData:isDerivedFrom("SQLStatement") .OR. xData:isDerivedFrom("SQLDataSet")  // Sql-express
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_SQLEXPRESS
   ELSEIF xData:isDerivedFrom("dsACEServer")    // ACE Server
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_ACESERVER
   ELSEIF xData:isDerivedFrom("SpSql")    // SpSql silex
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_SPSQL //
   ELSEIF xData:isDerivedFrom("MyResult")    // MYSQL
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_MYSQL
   ELSE                                                             // Database object
     aGetListItem[nGETLIST_SUBTYPE] := BROWSE_DATAOBJECT
   ENDIF
ELSEIF Valtype(xData) = 'N'
  aGetListItem[nGETLIST_SUBTYPE] := BROWSE_ADSHANDLE
ENDIF

::dataSource      := xData
::dataSourceType  := aGetListItem[nGETLIST_SUBTYPE]                           // J. Duijf 27-11-2002

IF Valtype(aGetListItem[aGETLIST_COLOR]) = 'A' .AND. ;
           Valtype(aGetListItem[aGETLIST_COLOR,1]) = 'B'
  bColorBlock := aGetListItem[aGETLIST_COLOR,1]

ELSEIF Valtype( aColor ) = 'A' .AND. Valtype(aGetListItem[aGETLIST_COLOR]) # 'B'
  bColorBlock := &('{|| {' + DC_XtoC(aColor[1]) + ',' + DC_XtoC(aColor[2]) + '}}' )

ELSEIF Valtype(aGetListItem[aGETLIST_COLOR]) = 'B'
  bColorBlock := aGetListItem[aGETLIST_COLOR]

ELSE

  bColorBlock := nil

ENDIF

IF Empty(aColor)
  aColor := { nil,15 }
ENDIF

DEFAULT aArray[9] := .f.,     ; //
        aArray[10] := nil,    ; // Thumb records
        aArray[11] := 1,      ; // Header Lines
        aArray[12] := ';',    ; // Header Delimiter
        aArray[13] := 1,      ; // Footer Lines
        aArray[14] := ';',    ; // Footer Delimiter
        aArray[15] := nil,    ; // Header Presentation Array
        aArray[16] := nil,    ; // Footer Presentation Array
        aArray[17] := .f.,    ; // No vertical scroll
        aArray[18] := .f.,    ; // No horizontal scroll
        aArray[19] := .f.,    ; // No Soft Track
        aArray[20] := .f.,    ; // Fit
        aArray[21] := .f.,    ; // Optimize
        aArray[24] := .f.,    ; // No Size Cols
        aArray[31] := .f.,    ; // Descending
        aArray[32] := 0,      ; // refresh time
        aArray[33] := .f.,    ; // Right Button Select
        aArray[34] := .f.,    ; // Don't disable descending toggle
        aArray[35] := nil,    ; // Code block to evaluate when Right-Button clicked in Data Area
        aArray[36] := nil,    ; // Array Filter
        aArray[37] := .F.,    ; // No Restore if .T.
        aArray[40] := .f.,    ; // Resize Columns
        aArray[41] := .f.,    ; // Tag Mode
        aArray[42] := 0,      ; // Tag Element
        aArray[43] := nil,    ; // Tag Foreground Color
        aArray[44] := GRA_CLR_YELLOW, ; // Tag Background Color
        aArray[45] := 0,      ; // Tag Background Color
        aArray[46] := .f.,    ; // Use Visual Style
        aArray[47] := nil,    ; // Refresh When
        aArray[48] := nil,    ; // Refresh Block
        aArray[49] := nil,    ; // OwnerDraw Block
        aArray[50] := 0       ; // Max width for auto fit

::tagEnable    := aArray[41]
::tagElement   := aArray[42]
::tagColorFG   := aArray[43]
::tagColorBG   := aArray[44]
::tagMode      := aArray[45]
::lastTagRec   := 0
::lastTagPos   := 0
IF IsMemberVar(self,'USEVISUALSTYLE')
  ::useVisualStyle := aArray[46]
ENDIF
::CellEditMode := .f.
::CellEditSetup()
::resizeColumns := aArray[40]

aSort := DC_BrowseSort()
::sortColorSelectedFG := aSort[1]
::sortColorSelectedBG := aSort[2]
::sortColorNonselectableFG := aSort[3]
::sortColorNonselectableBG := aSort[4]
::sortImageAscend := aSort[5]
::sortImageDescend := aSort[6]
::sortColorUnSelectedFG := aSort[7]
::sortColorUnSelectedBG := aSort[8]

IF !Empty(aArray[25])
  ::sortColorSelectedFG := aArray[25]
ENDIF
IF !Empty(aArray[26])
  ::sortColorSelectedBG := aArray[26]
ENDIF
IF !Empty(aArray[27])
  ::sortColorUnselectedFG := aArray[27]
ENDIF
IF !Empty(aArray[28])
  ::sortColorUnselectedBG := aArray[28]
ENDIF
IF !Empty(aArray[29])
  ::sortImageAscend := aArray[29]
ENDIF
IF !Empty(aArray[30])
  ::sortImageDescend := aArray[30]
ENDIF
IF !Empty(aArray[38])
  ::sortColorNonselectableFG := aArray[38]
ELSEIF Empty(::sortColorNonselectableFG)
   //
   // J. Duijf, 10-03-2011
   // Check for defauls first in presentation parameters
   //  ::sortColorNonselectableFG := GRA_CLR_WHITE
   ::sortColorNonselectableFG := DC_GetPresentation(::browsePres,XBP_PP_COL_HA_FGCLR,GRA_CLR_WHITE)
ENDIF
IF !Empty(aArray[39])
  ::sortColorNonselectableBG := aArray[39]
ELSEIF Empty(::sortColorNonselectableBG)
   //
   // J. Duijf, 10-03-2011
   // Check for defauls first in presentation parameters
   // ::sortColorNonselectableBG := GRA_CLR_DARKGRAY
   ::sortColorNonselectableBG := DC_GetPresentation(::browsePres,XBP_PP_COL_HA_BGCLR,GRA_CLR_DARKGRAY)
ENDIF
IF !Empty(aArray[37])
  ::noRestore := aArray[37]
ENDIF
DEFAULT ::sortColorUnselectedFG := ::sortColorNonselectableFG
DEFAULT ::sortColorUnselectedBG := ::sortColorNonselectableBG
//
// J. Duijf 10-03-2011
// If there is NO colour block. Test presentation parameters for
// default zebra striping
//

if ValType(bColorBlock) <> "B"

   if Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_ENABLED,0) > 0          // Zebra enabled
      //
      // If zebra and no zebra colors are defined, use default Black/White + Black/Yellow
      // It's better to add these colors to DC_BrowPres() in _dcpres.prg
      //
      //
      aEvenColor           := nil
      aOddColor            := nil
      if Ascan(::browsePres,{|a|a[1] == XBP_PP_COL_ZEBRA_ODDROW_FGCLR}) > 0
         aEvenColor        := {Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_ODDROW_FGCLR,GRA_CLR_WHITE),;
                               Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_ODDROW_BGCLR,GRA_CLR_YELLOW)}
      endif
      if Ascan(::browsePres,{|a|a[1] == XBP_PP_COL_ZEBRA_EVENROW_FGCLR}) > 0
         aOddColor         := {Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_EVENROW_FGCLR,GRA_CLR_BLACK),;
                               Dc_GetPresentation(::browsePres,XBP_PP_COL_ZEBRA_EVENROW_BGCLR,GRA_CLR_WHITE)}
      endif
      bColorBlock    := {||::ZebraColor(aEvenColor,aOddColor)}
   endif
endif

/*
if ValType(bColorBlock) = "B"
   //
   // J. Duijf 11-03-2011
   // Workaround for PDR 6365
   // The colorBlock should return nil or a 4 element array to prevent an incorrect
   // Hilite forground color.
   // {CellFGColor,CellBGColor,CellHiliteFGColor,CellHiliteBGColor}
   // The highlite colors come from the presentation parameters.
   // If undefined, the default BG/FG hilite colors are BLUE/WHITE
   //
   bTmpColorBlock    := bColorBlock
   bColorBlock       := {|a|a := Eval(bTmpColorBlock,self),;
                              iif(ValType(a) = "A" .and. Len(a) = 2,;
                                 (Aadd(a,Dc_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_FGCLR,GRA_CLR_WHITE)), ;
                                  Aadd(a,Dc_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_BGCLR,GRA_CLR_BLUE))),nil),a}
endif
*/

::sortColumn := 0
::descending := aArray[31]
::descendToggle := !aArray[34]
::oHasFocus := NIL
::dataLink := xData
::outputBlock := DC_GetBlock(aGetListItem[bGETLIST_VAR])
::inputBlock := DC_GetBlock(aGetListItem[bGETLIST_ACTION])
::whenBlock := DC_GetBlock(aGetListItem[bGETLIST_WHEN])
::hideBlock := DC_GetBlock(aGetListItem[bGETLIST_HIDE])
::protectBlock := DC_GetBlock(aGetListItem[bGETLIST_PROTECT])
::mark := aArray[1]
::markColorBlock := aArray[3]
::rowColorBlock := _RowColor( self, aArray, aColor, xData, bColorBlock )
::freezeLeft := aArray[7]
::freezeRight := aArray[8]
::isScope := aArray[9]
::font := aGetListItem[xGETLIST_OPTIONS2,4]
::pointerBlock := DC_GetBlock(aGetListItem[xGETLIST_OPTIONS6])
::thumbRecords := aArray[10]
::autoLock := !(aGetListItem[xGETLIST_OPTIONS2,6] .OR. Valtype(xData)#'C')
::headerLines := aArray[11]
::headerDelim := aArray[12]
::footerLines := aArray[13]
::footerDelim := aArray[14]
::headerPres := aArray[15]
::footerPres := aArray[16]
::columns := {}
::fit := aArray[20]
::optimize := aArray[21]
::arrayElement := 1
::refreshInterval := aArray[32]
::refreshWhenBlock := aArray[48]
::refreshBlock := aArray[47]
::rightButtonSelect := aArray[33]
::hasHeader := .f.
::hasFooter := .f.
::rightButtonDown := aArray[35]
::filter := aArray[36]
::ownerDrawBlock := aArray[49]
::fitMax := aArray[50]

#if XPPVER > 02000000
IF Valtype(aArray[52]) == 'N'
  ::navigationMode := aArray[52]
ELSE
  ::navigationMode := XBPBRW_NAVIGATION_1XCOMPATIBLE
ENDIF
#endif

::recordObject := aArray[53]

IF Valtype(::recordObject) == 'O'
  ::navigateBlock := {||(::dataSource)->(DC_DbScatter(::recordObject))}
  Eval(::navigateBlock,::recordObject)
ELSE
  ::navigateBlock := {||nil}
ENDIF

IF Valtype(::ownerDrawBlock) == 'B'
  ::drawMode := XBP_DRAW_OWNER
ENDIF

::cargo := { ::getListPointer, ;
             aGetList, ;
             aGetListItem[xGETLIST_CARGO], ;
             ::arrayElement, ;
             ::dataSource }

//
// J. Duijf 12-04-2011 Move up in this code
//
//xFont := aGetListItem[cGETLIST_FONT]
//IF Empty(xFont)
//  DEFAULT xFont := ::config:font
//ENDIF

//IF !Empty(xFont)
//  IF Valtype(xFont) == 'B'
//    xFont := Eval(xFont)
//  ENDIF
//  ::setFontCompoundName(Alltrim(xFont))
//ENDIF

::vScroll   := !aArray[17]
::hScroll   := !aArray[18]
::softTrack := !aArray[19]
::sizeCols  := !aArray[24]

::itemSelected := ;
         { || IIF(Valtype(::dataSource)='B',Eval(::dataSource,self),nil), ;
              IIF(Valtype(::mark)='N',::dataSource[::arrayElement,::mark] := ;
                           !::dataSource[::arrayElement,::mark],nil), ;
              IIF(Valtype(::mark)='B',Eval(::mark,self),nil), ;
              IIF(Valtype(::pointerBlock)=='B',Eval(::pointerBlock,::arrayElement),nil),;
              ::RefreshCurrent() }

IF Valtype(xOptions[23]) = 'B'
  ::itemSelected := DC_MergeBlocks(::itemSelected,xOptions[23])
ELSEIF Valtype(bBlock:=DC_GetBlock(aGetListItem[bGETLIST_ACTION])) = 'B'
  ::itemSelected := DC_MergeBlocks(::itemSelected,bBlock)
ENDIF

IF Valtype(::pointerBlock) == 'B'
  ::itemMarked := {||Eval(::pointerBlock,::arrayElement)}
ENDIF

IF Valtype(xOptions[22]) = 'B'
  ::itemMarked := DC_MergeBlocks(::itemMarked,xOptions[22])
  ::isItemMarkedCodeBlock := .t.
ENDIF

IF DC_GetWhenOnBrowseItemMarked()
  ::isItemMarkedCodeBlock := .t.
ENDIF

IF Valtype(xData) = 'C'
  xData := Alltrim(Upper(xData)) + '->'
ENDIF

IF Valtype(::thumbRecords) == 'B'
  lThumbLock := Eval(::thumbRecords)
ELSEIF Valtype(::dataSource) == 'C'
  DEFAULT ::thumbRecords := 32000
  lThumbLock := ::thumbRecords == 0 .OR. (::dataSource)->(RecCount()) > ::thumbRecords
ELSE
  lThumbLock := .f.
ENDIF

IF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_ARRAY // array browse

  ::skipBlock     := {|n,o|DC_SkipArray(n,self)}

  ::goTopBlock    := {|o| ::arrayElement := ;
                           ArrayGoTop(self), ;
                           ::cargo[4] := ::arrayElement}

  ::goBottomBlock := {|o| ::arrayElement := ;
                           ArrayGoBottom(self), ;
                           ::cargo[4] := ::arrayElement}

  ::posBlock      := {|o| IIF(::descending,Len(::dataSource)-::arrayElement,::arrayElement)}
  ::phyPosBlock   := {| | ::arrayElement }

  ::lastPosBlock  := {||Len(::dataSource)}
  ::firstPosBlock := {||1}

  nPointer := aGetListItem[xGETLIST_OPTIONS5]
  IF Empty(nPointer) .AND. Valtype(::pointerBlock) == 'B'
    nPointer := Eval(::pointerBlock)
  ENDIF

  IF !Empty(::filter)
    ::arrayElement := ArrayGoTop(self)
  ELSEIF Valtype(nPointer) = 'N' .AND. nPointer > 0 .AND. nPointer <= Len(xData)
    ::arrayElement := nPointer
  ELSEIF ::descending
    ::arrayElement := Len(::dataSource)
  ENDIF

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_SQLEXPRESS // SQLExpress browse

  ::skipBlock     := {|n,o| o:dataSource:Skipper(n) }
  ::goTopBlock    := {|o| o:dataSource:goTop() }
  ::goBottomBlock := {|o| o:dataSource:goBottom() }
  ::phyPosBlock   := {|o| o:dataSource:RecNo() }
  // Navigation blocks for VScroll bar
  ::posBlock      := {|o| o:dataSource:RecNo() }
  ::lastPosBlock  := {|o| o:dataSource:RecCount() }
  ::firstPosBlock := {|| 1 }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_MYSQL // SQLExpress browse

  ::skipBlock     := {|n,o| o:dataSource:Skipper(n) }
  ::goTopBlock    := {|o| o:dataSource:goTop() }
  ::goBottomBlock := {|o| o:dataSource:goBottom() }
  ::phyPosBlock   := {|o| o:dataSource:RecNo() }
  // Navigation blocks for VScroll bar
  ::posBlock      := {|o| o:dataSource:RecNo() }
  ::lastPosBlock  := {|o| o:dataSource:RecCount() }
  ::firstPosBlock := {|| 1 }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_DATAOBJECT // Database object  J.Duijf 27-11-2002

  ::skipBlock     := {|n|::dataSource:DbSkipper(n)    }
  ::goTopBlock    := {| |::dataSource:DbGoTop()       }
  ::goBottomBlock := {| |::dataSource:DbGoBottom()    }
  ::phyPosBlock   := {| |::dataSource:Recno()         }
  ::GoPosBlock    := {|n|::dataSource:DbGoPosition(n) }
  ::posBlock      := {| |::dataSource:DbPosition()    }
  ::lastPosBlock  := {||100}
  ::firstPosBlock := {|| 1 }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_ACESERVER // ACEServer++

   ::skipBlock     := {|n| ACEDbSkipper(n, ::dataSource) }
   ::goTopBlock    := {| | ::dataSource:GoTop()          }
   ::goBottomBlock := {| | ::dataSource:GoBottom()       }
   ::phyPosBlock   := {| | ::dataSource:Recno()          }
   ::posBlock      := {| | ::dataSource:Position()*10    }
   ::goPosBlock    := {|n| ::dataSource:GoPosition(n/10) }
   ::lastPosBlock  := {| | 1000 }
   ::firstPosBlock := {| | 0    }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_SPSQL // SPSQL silex

   ::skipBlock     := {|n| ::dataSource:SQLViewSkip(n) }
   ::goTopBlock    := {| | ::dataSource:GoTop()          }
   ::goBottomBlock := {| | ::dataSource:GoBottom()       }
 //  ::phyPosBlock   := {| | ::dataSource:dataarea:Recno()          }
  ::phyPosBlock   := {| |::dataSource:getNum()}
  ::posBlock      := {| |::dataSource:getNum()}

//    ::posBlock      := {| |posicionFisica(::dataSource:hIndex)}
  //  ::phyPosBlock   := {| |posicionFisica(::dataSource:hIndex)}
 //  ::posBlock      := {| | ::dataSource:dataarea:Recno() }

   ::goPosBlock    := {|n| ::dataSource:dataarea:goTo(::dataSource:getNum()) }

   ::lastPosBlock  := {| | ::dataSource:LastRec() }
   ::firstPosBlock := {| | 1 }

ELSEIF IsFunction('AceWBlock')

  // ** NOTE **
  // since macro compiled codeblocks are "late bound", they require the ACE Wrapper
  // functions to be passed exactly as they were defined in ACE.PRG (in most cases
  // with a leading underscore "_" character). (Boris - 2000/06/05)

  xData := StrTran(xData,"->","")
  ::skipBlock     := &( '{|n|AceWBlock("' + xData + '", {|x|_DbSkipper(x)},n)}' )
  ::goTopBlock    := &( '{| |AceWBlock("' + xData + '", {||_DbGoTop()})}' )
  ::goBottomBlock := &( '{| |AceWBlock("' + xData + '", {||_DbGoBottom()})}' )
  ::phyPosBlock   := &( '{| |AceWRecNo("' + xData + '")}' )

  // scrollbar navigation
  if &('AceWBlock("' + xData + '", {||OrdNumber()})') == 0
    ::posBlock      := &( '{| |AceWBlock("' + xData + '", {||_RecNo()})}' )
    ::goPosBlock    := &( '{|n|AceWBlock("' + xData + '", {|x|_DbGoTo(x)},n)}' )
    ::lastPosBlock  := &( '{| |AceWBlock("' + xData + '", {||_LastRec()})}' )
    ::firstPosBlock := {|| 1}
  else
    ::posBlock      := &( '{| |AceWBlock("' + xData + '", {||_DbPosition()})}' )
    ::goPosBlock    := &( '{|n|AceWBlock("' + xData + '", {|x|_DbGoPosition(x)},n)}' )
    ::lastPosBlock  := {|| 100}
    ::firstPosBlock := {|| 0 }
  endif

ELSEIF IsFunction('XPPSETSCOPE')

  ::firstPosBlock := {||1}
  ::LastPosBlock  := {||100}
  ::PosBlock      := {||(::dataSource)->(DbPosition(,::descending))}
  ::GoPosBlock    := {|n|(::dataSource)->(DbGoPosition(n,::descending))}
  ::skipBlock     := {|n|(::dataSource)->(DbSkipper(iif(::descending, n*(-1), n)))}
  ::goTopBlock    := {||(::dataSource)->(IIF(::descending,DbGoBottom(),DbGoTop()))}
  ::goBottomBlock := {||(::dataSource)->(IIF(::descending,DbGoTop(),DbGoBottom()))}
  ::phyPosBlock   := {||(::dataSource)->(Recno())}

ELSEIF ::isScope .AND. DC_ScopeExpress()

  ::skipBlock     := {|n|(::dataSource)->(DC_DbSkipper(n)) }
  ::goTopBlock    := {| |(::dataSource)->(DC_DbGoTop())    }
  ::goBottomBlock := {| |(::dataSource)->(DC_DbGoBottom()) }
  ::phyPosBlock   := {|n|IIF(Valtype(n)='N',(::dataSource)->(dbGoTo(n)), ;
                         (::dataSource)->(Recno()))}
  IF lThumbLock
    ::posBlock      := {||50}
    ::lastPosBlock  := {||100}
  ELSE
    ::posBlock      := {| |(::dataSource)->(DC_KeyNo32k())   }
    ::lastPosBlock  := {| |(::dataSource)->(DC_KeyCount32k())}
  ENDIF
  ::firstPosBlock := {| | 1 }

ELSEIF aGetListItem[nGETLIST_SUBTYPE] = BROWSE_ADSHANDLE // Advantage Server Handle

  ::firstPosBlock := {||1}
  ::lastPosBlock  := {||100}
  ::goTopBlock    := {||AdsGoToTop(::dataSource)}
  ::goBottomBlock := {||AdsGoToBottom(::dataSource)}
  ::skipBlock     := {|n|DC_AdsSkipper(::dataSource,n)}
  ::phyPosBlock   := {|n|n:=0,AdsGetRecordNum(::dataSource,ADS_IGNOREFILTERS,@n),n}
  ::GoPosBlock    := {|n|DC_AdsGoPosition(::dataSource,n)}
  ::posBlock      := {|n|DC_AdsPosition(::dataSource )}

ELSE

  IF lThumbLock
    ::posBlock      := {||50}
  ELSE
    ::PosBlock      := {||(::dataSource)->(DC_DbPosition(,::descending))}
  ENDIF
  ::firstPosBlock := {||1}
  ::lastPosBlock  := {||100}
  ::GoPosBlock    := {|n,x|x := (::dataSource)->(DC_DbGoPosition(n,::descending)),Eval(::navigateBlock),x}
  ::skipBlock     := {|n,x|x := (::dataSource)->(DC_DbSkipper(n,,::descending,::skipUnique)),Eval(::navigateBlock),x}
  ::goTopBlock    := {|  x|x := (::dataSource)->(IIF(::descending,DC_DbGoBottom(),DC_DbGoTop())),Eval(::navigateBlock),x}
  ::goBottomBlock := {|  x|x := (::dataSource)->(IIF(::descending,DC_DbGoTop(),DC_DbGoBottom())),Eval(::navigateBlock),x}
  ::phyPosBlock   := {|n  |IIF(Valtype(n)='N',(::dataSource)->(dbGoTo(n)),(::dataSource)->(Recno()))}

ENDIF

// ::headerRbDown := {|a,n,o|_HeaderRbDown( a,n,o,self )}

IF !Empty(::getList:getOptions[nGETOPT_HILITECOLOR])
  ::killInputFocus := {|x,y,o|  ;
                        DC_HiliteGets(o,2,o:getList:getOptions[nGETOPT_HILITECOLOR])}
  ::setInputFocus := {|x,y,o| ;
                       DC_HiliteGets(o,1,o:getList:getOptions[nGETOPT_HILITECOLOR])}
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_GROUP],'O')
IF Valtype( bBlock ) = 'B'
  Eval( bBlock, self )
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_PREEVAL])
IF Valtype(bBlock) = 'B'
  Eval( bBlock, self )
ENDIF

cID := aGetListItem[cGETLIST_ID]

IF (!Empty(DC_AutoRestoreBrowse()) .AND. !Empty(cID) .AND. !::noRestore) .OR. ;
      !Empty(::restoreConfig)

  IF Valtype(::restoreConfig) = 'C'
    aConfig := DC_ARestore(::restoreConfig)
  ELSEIF Valtype(::restoreConfig) = 'A'
    aConfig := ::restoreConfig
  ELSE
    IF DC_AutoRestoreBrowse()[3]
      ::restoreProc := ProcName(3)
      IF 'READGUI' $ ::restoreProc
        ::restoreProc := ProcName(4)
      ENDIF
    ELSE
      ::restoreProc := ''
    ENDIF

    ::restoreProc += ':' + DC_XtoC(cID)

    IF Valtype(DC_AutoRestoreBrowse()[1]) == 'B'
      aConfig := Eval(DC_AutoRestoreBrowse()[1],::restoreProc)
    ELSE
      aConfig := DC_RegQuery(DC_AutoRestoreBrowse()[1],DC_AutoRestoreBrowse()[2], ;
                             ::restoreProc)
    ENDIF
  ENDIF

  IF Valtype(aConfig) = 'A'

    nPointer := 1
    nColumns := 0
    ::defaultColumnsGetList := {}
    ASize(aConfig,7)

    DO WHILE aGetList[::getListPointer+nPointer,nGETLIST_TYPE] $ { GETLIST_BROWSECOL, GETLIST_SETPARENT } .AND. ;
             ::getListPointer+nPointer <= Len(aGetList)
      IF aGetList[::getListPointer+nPointer,nGETLIST_TYPE] == GETLIST_BROWSECOL
        nColumns++
      ENDIF
      nPointer++
    ENDDO

    BEGIN SEQUENCE

    IF nColumns # Len(aConfig[3])
      IF Valtype(::configMessageBlock) == 'B'
        Eval(::configMessageBlock,self)
      ELSE
        DCMSGBOX 'Your Saved Browse configuration cannot be restored:', ;
                  DC_AutoRestoreBrowse()[2]+'\'+::restoreProc ;
          TIMEOUT 5

        IF Valtype(DC_AutoRestoreBrowse()[1]) == 'B'
          Eval(DC_AutoRestoreBrowse()[1],::restoreProc)
        ELSE
          DC_RegWrite(DC_AutoRestoreBrowse()[1],DC_AutoRestoreBrowse()[2], ;
                     ::restoreProc, 0)
        ENDIF
      ENDIF
      BREAK
    ENDIF

    nPointer := 1
    DO WHILE aGetList[::getListPointer+nPointer,nGETLIST_TYPE] == GETLIST_BROWSECOL .AND. ;
       ::getListPointer+nPointer <= Len(aGetList)
      nFound := AScan(aConfig[3],{|a|a[1]=nPointer})
      AAdd(::defaultColumnsGetList,AClone(aGetList[::getListPointer+nPointer]))
      aGetList[::getListPointer+nPointer,cGETLIST_LEVEL] := { nFound, nPointer }
      IF nFound > 0
        aGetList[::getListPointer+nPointer,nGETLIST_WIDTH] := aConfig[3,nFound,2,1]
        aGetList[::getListPointer+nPointer,lGETLIST_PIXEL] := .T.
        ASize(aConfig[3,nFound],3)
        IF !Empty(aConfig[3,nFound,3])
          aGetList[::getListPointer+nPointer,cGETLIST_FONT] := Trim(aConfig[3,nFound,3])
        ENDIF
        IF !Empty(aConfig[4]) .OR. !Empty(aConfig[5])
          DEFAULT aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS2] := {}
          ASize(aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS2],4)
          aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS2][3] := aConfig[4]
          aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS2][4] := aConfig[5]
        ENDIF
        IF !Empty(aConfig[6]) .OR. !Empty(aConfig[7])
          DEFAULT aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS8] := {}
          ASize(aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS8],4)
          aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS8][3] := aConfig[6]
          aGetList[::getListPointer+nPointer,xGETLIST_OPTIONS8][4] := aConfig[7]
        ENDIF
      ELSE
        // aGetList[::getListPointer+nPointer,bGETLIST_HIDE] := {||.t.}
      ENDIF
      nPointer++
    ENDDO
    ASort( aGetList, ::getListPointer+1, nPointer-1, ;
            {|a,b|a[cGETLIST_LEVEL,1] < b[cGETLIST_LEVEL,1]} )

    END SEQUENCE
  ENDIF
ENDIF

::scrollBarTipBlock := aGetListItem[xGETLIST_OPTIONS9]
IF Valtype(::scrollBarTipBlock) == 'B'
  ::getList:isDataToolTips := .t.
ENDIF

IF ::tagEnable
  ::keyboard := DC_MergeBlocks(::keyboard,{|n|::Tag(,,n)} )
ENDIF
RETURN self

* ------------------

METHOD DC_XbpBrowse:create( oParent, oOwner, aPos, aSize, aPres, lVisible )

LOCAL aChildren, i

::XbpBrowse:create( oParent, oOwner, aPos, aSize, aPres, lVisible )

DEFAULT oParent := ::setParent()

IF Valtype(::hideBlock)='B' .AND. Eval(::hideBlock)
  ::hide()
ENDIF

::DC_XbaseParts:setOrig(self)

aChildren := ::childList()
FOR i := 1 TO Len(aChildren)
  IF aChildren[i]:isDerivedFrom("XbpScrollBar")
    IF aChildren[i]:type == XBPSCROLL_VERTICAL .AND. Valtype(::scrollBarTipBlock) == 'B'
      aChildren[i]:cargo :=  {{||.t.},::scrollBarTipBlock,'',0,0,0,nil,::dataSource}
      // :cargo[1] = Enable Code Block
      // :cargo[2] = Return Code Block
      // :cargo[3] = Value to display in tooltip
      // :cargo[4] = Last moused cell record offset
      // :cargo[5] = Last selected record
      // :cargo[6] = Last moused cell record
      // :cargo[7] = Pointer to Browse object
      // :cargo[8] = Alias(), Array or SQLExpress object
    ENDIF
    IF IsMemberVar(aChildren[i],"isVisualStyle")
      aChildren[i]:useVisualStyle := ::config:useVisualStyleScrollBars
    ENDIF
  ENDIF
NEXT

RETURN self

* --------------------

METHOD DC_Xbpbrowse:right()

::XbpBrowse:right()

/*
IF nColumn == ::colPos
  oColumn := ::getColumn(nColumn)
  nGetListPos := oColumn:getListPointer
  nFound := AScan(::getList:getListArray,;
                  {|a,b|b:=DC_GetBlock(a[bGETLIST_PARENT]),;
                  a[nGETLIST_TYPE]==GETLIST_BROWSECOL .AND. ;
                  Eval(b)==self}, nGetListPos+1)
  IF nFound > 0
    ::getList:getListPointer := nFound

// DCQDEBUG self, ::getList:getListPointer

    oXbp := DC_XbpColumn():new( self, ::getList, .t. )

// DCQDEBUG oXbp, ::colCount

  ENDIF
ENDIF
*/

RETURN self

* --------------------

METHOD DC_XbpBrowse:customDrawCell( oPS, aInfo )

IF !Empty(::ownerDrawBlock)
  RETURN Eval( ::ownerDrawBlock, oPS, aInfo, self )
ENDIF

RETURN .t.

* --------------------

METHOD DC_XbpBrowse:rowColor()

LOCAL aColor, aBrowseColor := DC_BrowseColor()

IF Valtype(::dataSource) == 'C'
  IF ::tagEnable .AND. (::dataSource)->(DC_Tagged())
    RETURN {::tagColorFG, ::tagColorBG}
  ELSEIF (::dataSource)->(RecNo()) <> ::nBRow
    ::lBRow := !::lBRow
    ::nBRow := (::dataSource)->(RecNo())
  ENDIF
ELSEIF Valtype(::dataSource) == 'A' .AND. Len(::dataSource) > 0
  IF ::tagEnable .AND. ::arrayElement > 0 .AND. ::dataSource[::arrayElement,::tagElement]
    RETURN {::tagColorFG, ::tagColorBG}
  ELSEIF (::arrayElement) % 2 == 0
    ::lBRow := .t.
  ELSE
    ::lBrow := .f.
  ENDIF
ELSE
  ::lBrow := .t.
ENDIF

IF Valtype(aBrowseColor) == 'A'
  IF ::lBRow
    aColor := {aBrowseColor[1],aBrowseColor[2]}
  ELSE
    aColor := {aBrowseColor[3],aBrowseColor[4]}
  ENDIF
ELSE
  aColor := { nil, nil}
ENDIF

RETURN aColor

* --------------------

METHOD DC_XbpBrowse:destroy()

LOCAL aConfig

IF Valtype(::timer) == 'O' .AND. ::timer:isDerivedFrom('DC_SetTimerEvent')
  ::timer:destroy()
  Sleep(10)
ENDIF

IF !Empty(DC_AutoRestoreBrowse()) .AND. !Empty(::getListPointer) .AND. ;
   !Empty(::getList:getListArray) .AND. !::noRestore .AND. !Empty(::restoreProc)

  aConfig := ::save()
  IF Valtype(DC_AutoRestoreBrowse()[1]) == 'B'
    Eval(DC_AutoRestoreBrowse()[1],::restoreProc,aConfig)
  ELSE
    DC_RegWrite(DC_AutoRestoreBrowse()[1],DC_AutoRestoreBrowse()[2], ;
                ::restoreProc, aConfig)
  ENDIF
ENDIF

::Xbpbrowse:destroy()

//
// J. Duijf 12-04-2011
// Remove references (if any)
//
/*
::Datalink        := nil
::DataSource      := nil
::Cargo           := nil
::GetList         := nil
::xbp             := nil
::config          := nil
::GoBottomBlock   := nil
::GoPosBlock      := nil
::PhyPosBlock     := nil
::PosBlock        := nil
::LastPosBlock    := nil
::GoTopBlock      := nil
::SkipBlock       := nil
::HitBottomBlock  := nil
::HitTopBlock     := nil
::ItemSelected    := nil
::LastPosBlock    := nil
::StableBlock     := nil
::KillInputFocus  := nil
::SetInputFocus   := nil
::BrowsePres      := nil
::ItemSelected    := nil
::StableBlock     := nil
::KillInputFocus  := nil
::OrigFont        := nil
::Cargo           := nil
::Timer           := nil
*/

RETURN self

* -----------------

METHOD DC_XbpBrowse:restoreDefaults()

LOCAL i, aColumns[::colCount], GetList

IF Empty(::defaultColumnsGetList)
  RETURN self
ENDIF

::lockUpdate(.t.)

FOR i := 1 TO Len(aColumns)
  ::delColumn(1)
NEXT

GetList := ::defaultColumnsGetList

DCREAD GUI PARENT self EXIT SAVE

::lockUpdate(.f.)

RETURN self

* --------------------

METHOD DC_XbpBrowse:fitColumns()

LOCAL nBrowseWidth, aChildList, i, nDelta, nColWidth, oPS, oColumn

oPS := ::lockPs()
nBrowseWidth := ::currentSize()[1]
aChildList := ::childList()
FOR i := 1 TO Len(aChildList)
  IF aChildList[i]:isDerivedFrom('XbpScrollBar') .AND. ;
    aChildList[i]:type == XBPSCROLL_VERTICAL
    nBrowseWidth -= aChildList[i]:currentSize()[1]
  ENDIF
NEXT

nColWidth := 0
FOR i := 1 TO ::colCount
  nColWidth += ::getColumn(i):currentSize()[1]
NEXT

nDelta := (nBrowseWidth-nColWidth)/::colCount
FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  oColumn:setSize({oColumn:currentSize()[1]+nDelta,oColumn:currentSize()[2]})
NEXT
::unLockPs(oPS)
::setSize(::currentSize())

RETURN nil

* --------------------

METHOD DC_XbpBrowse:resizeColumns( aOldSize, aNewSize )

LOCAL i, nDelta, oPS, oColumn, aColumns[0]

IF ::xbpBrowse:status() = 0
  RETURN nil
ENDIF

oPS := ::lockPs()

FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  IF !oColumn:noAutoResizeColumn
    AAdd(aColumns,oColumn)
  ENDIF
NEXT

nDelta := Int((aNewSize[1]-aOldSize[1])/Len(aColumns))
FOR i := 1 TO Len(aColumns)
  oColumn := aColumns[i]
  oColumn:setSize({oColumn:currentSize()[1]+nDelta,oColumn:currentSize()[2]})
NEXT
::unLockPs(oPS)
::setSize(::currentSize())

RETURN nil

* --------------------

METHOD DC_XbpBrowse:refreshView()

LOCAL nRows := ::rowPos-1, i

FOR i := 1 TO nRows
  Eval(::skipBlock,-1)
NEXT
::refreshAll()
FOR i := 1 TO nRows
  Eval(::skipBlock,1)
NEXT

RETURN nil

* --------------------

METHOD DC_XbpBrowse:handleEvent( nEvent, mp1, mp2 )

IF nEvent == DCGUI_EVENT_BROWSE_REFRESH
  IF ::isVisible()
    IF Empty(::refreshWhenBlock) .OR. Eval(::refreshWhenBlock)
      IF Valtype(::refreshBlock) == 'B'
        Eval(::refreshBlock,self)
      ELSE
        ::Refresh()
      ENDIF
    ENDIF
  ENDIF
ELSE
  RETURN ::XbpbrowseHandleEvent( nEvent, mp1, mp2 )
ENDIF
RETURN self

* --------------------

METHOD DC_XbpBrowse:setTimerRefresh()

DEFAULT ::refreshInterval := 0

IF ::refreshInterval > 0
  ::timer := DC_SetTimerEvent():new(::refreshInterval, ;
              {||PostAppEvent(DCGUI_EVENT_BROWSE_REFRESH,,,self)})
ENDIF

RETURN self

* --------------------

METHOD DC_XbpBrowse:optimize()

LOCAL i, oHeading

FOR i := 1 TO ::colCount
  oHeading := ::getColumn(i):heading
  _Optimize(nil,nil,oHeading)
NEXT

RETURN self

* --------------------

METHOD DC_XbpBrowse:dataDisplay( lEnable )

LOCAL nRecNo

IF Empty(::navBlocks)
  ::navBlocks := { ;
    ::posBlock, ;
    ::firstPosBlock, ;
    ::lastPosBlock, ;
    ::goPosBlock, ;
    ::skipBlock, ;
    ::goTopBlock, ;
    ::goBottomBlock, ;
    ::phyPosBlock, ;
    ::arrayElement }
ENDIF

IF lEnable
  ::posBlock := ::navBlocks[1]
  ::firstPosBlock := ::navBlocks[2]
  ::lastPosBlock := ::navBlocks[3]
  ::goPosBlock := ::navBlocks[4]
  ::skipBlock := ::navBlocks[5]
  ::goTopBlock := ::navBlocks[6]
  ::goBottomBlock := ::navBlocks[7]
  ::phyPosBlock := ::navBlocks[8]
  ::arrayElement := ::navBlocks[9]
  ::dataEnabled := .t.
  ::refreshAll()
ELSE
  ::posBlock := {||0}
  ::firstPosBlock := {||0}
  ::lastPosBlock  := {||0}
  ::GoPosBlock    := {||0}
  ::skipBlock     := {||0}
  ::goTopBlock    := {||0}
  ::goBottomBlock := {||0}
  ::phyPosBlock   := {||0}
  IF Valtype(::dataSource) == 'C'
    nRecNo := (::dataSource)->(RecNo())
    (::dataSource)->(dbGoBottom())
    (::dataSource)->(dbSkip())
    ::refreshAll()
    (::dataSource)->(dbGoto(nRecNo))
  ELSEIF Valtype(::dataSource) == 'A'
    ::navBlocks[9] := ::arrayElement
    ::arrayElement := 0
    ::refreshAll()
  ENDIF
  ::dataEnabled := .f.
ENDIF

RETURN self

* --------------------

METHOD DC_XbpBrowse:refresh()

LOCAL i, aEvents[0], nEvent, mp1, mp2, oXbp, nRecNo

IF !::dataEnabled
  RETURN self
ENDIF

if ::dataSourceType = BROWSE_DATABASE                             // J. Duijf 27-11-2002
   nRecNo   := (::dataSource)->(RecNo())                          // J. Duijf 27-11-2002
elseif ::dataSourceType = BROWSE_DATAOBJECT                       // J. Duijf 27-11-2002
   nRecno   := ::dataSource:Recno()                               // J. Duijf 27-11-2002
endif                                                             // J. Duijf 27-11-2002

PostAppEvent(xbeP_User+10,12,13,self)

IF Valtype(::freezeLeft) = 'A' .AND. Empty(::setLeftFrozen())
  ::setLeftFrozen( ::freezeLeft )
ENDIF
IF Valtype(::freezeRight) = 'A' .AND. Empty(::setRightFrozen())
  ::setRightFrozen( ::freezeRight )
ENDIF

::RefreshAll()

nEvent := -1
DO WHILE nEvent # 0
  nEvent := AppEvent(@mp1,@mp2,@oXbp,1)
  IF nEvent # 0 .AND. nEvent # xbeBRW_ItemMarked
    AAdd(aEvents,{nEvent,mp1,mp2,oXbp})
  ENDIF
ENDDO
FOR i := 1 TO Len(aEvents)
  PostAppEvent( aEvents[i,1], aEvents[i,2], aEvents[i,3], aEvents[i,4] )
NEXT

IF Valtype(nRecNo) == 'N'
   if ::dataSourceType = BROWSE_DATABASE                             // J. Duijf 27-11-2002
      (::dataSource)->(dbGoTo(nRecNo))
   elseif ::dataSourceType = BROWSE_DATAOBJECT                       // J. Duijf 27-11-2002
      ::dataSource:DbGoto(nRecno)                                    // J. Duijf 27-11-2002
   endif                                                             // J. Duijf 27-11-2002
ENDIF

RETURN self

* ---------------------------

METHOD DC_XbpBrowse:Resize( aOldSize, aNewSize )

::RefreshAll()
::InvalidateRect ( { 0, 0, aNewSize[1], aNewSize[2] } )

RETURN ::XbpStatic:Resize( aOldSize, aNewSize )

* ---------------------------

METHOD DC_XbpBrowse:Setsize( aSize )

LOCAL lStatus, bError := ErrorBlock({|e|Break(e)})

BEGIN SEQUENCE
lStatus := ::XbpBrowse:setSize(aSize)
END SEQUENCE

ErrorBlock(bError)

RETURN lStatus

* ---------------------------

METHOD DC_XbpBrowse:RefreshAll()

LOCAL nRecno := 0, lThumbLock

IF !::dataEnabled .or. ::Status <> XBP_STAT_CREATE
  RETURN self
ENDIF

if ::dataSourceType = BROWSE_DATABASE                             // J. Duijf 27-11-2002
   IF Empty((::dataSource)->(dbScope(1))) .AND. !Empty((::dataSource)->(DC_SetScope(0)))
     nRecno := (::dataSource)->(Recno())             // Save current record
     (::dataSource)->(DC_DbSkip())                   // Next
     (::dataSource)->(DC_DbSkip(-1))                 // Previous
     if (::dataSource)->(Bof()) .or. (::dataSource)->(Eof()) .OR. ;
       (::dataSource)->(RecNo()) != nRecno           // Current record NOT within scope, or deleted
       (::dataSource)->(DC_DbGoTop())                // To first record
       ::GoTop()                                     // Same goes for xbpBrowse
     endif
   ENDIF
elseif ::dataSourceType = BROWSE_DATAOBJECT                          // J. Duijf 27-11-2002
   nRecno := ::dataSource:Recno()                                    // J. Duijf 27-11-2002
   ::dataSource:DbSkip()                                             // J. Duijf 27-11-2002
   ::dataSource:DbSkip(-1)                                           // J. Duijf 27-11-2002
   if ::dataSource:Bof() .or. ::dataSource:Eof() .or. ;              // J. Duijf 27-11-2002
      ::datasource:Recno()  != nRecno                                // J. Duijf 27-11-2002
//      ::dataSource:DbGoTop()                                         // J. Duijf 27-11-2002 ::Gotop() does same
      ::GoTop()                                                      // J. Duijf 27-11-2002
   endif
elseif ::dataSourceType = BROWSE_ARRAY                               // J. Duijf 26-03-2011 Force initialisation of nPointer
   ::arrayElement       := Min(::arrayElement,Len(::dataSource))     // J. Duijf 19-04-2015 Keep ::arrayElement in range
   If Len(::dataSource) > 0 .and. ::arrayElement = 0                 // J. Duijf 19-04-2015 Set nPointer to 1 if data availeble
      ::arrayElement    := 1
   Endif
   IIF(Valtype(::pointerBlock)=='B',Eval(::pointerBlock,::arrayElement),nil)
endif

IF Valtype(::thumbRecords) == 'B' .AND. Valtype(::dataSource) = 'C'
  lThumbLock := Eval(::thumbRecords)
  IF ::isScope .AND. DC_ScopeExpress()
    IF lThumbLock
      ::posBlock      := {||50}
      ::lastPosBlock  := {||100}
    ELSE
      ::posBlock      := {| |(::dataSource)->(DC_KeyNo32k())   }
      ::lastPosBlock  := {| |(::dataSource)->(DC_KeyCount32k())}
    ENDIF
  ELSE
    IF lThumbLock
      ::posBlock      := {||50}
    ELSE
      ::PosBlock      := {||(::dataSource)->(DC_DbPosition(,::descending))}
    ENDIF
  ENDIF
ENDIF

Return ::xbpBrowse:RefreshAll()

* ---------------------------

METHOD DC_XbpBrowse:WhenHideEval( lEnable, lShow, lProtect )

IF Valtype(::whenBlock) = 'B'
  lEnable:= Eval(::whenBlock,self)
  IF lEnable
    IF !::isEnabled()
      ::enable()
    ENDIF
  ELSE
    IF ::isEnabled()
      ::disable()
    ENDIF
  ENDIF
ENDIF

IF Valtype(::hideBlock) = 'B'
  lShow := !Eval(::hideBlock,self)
  IF lShow
    ::show()
  ELSEIF !lShow
    ::hide()
  ENDIF
ENDIF

IF Valtype(lProtect) == 'L' .AND. Valtype(::protectBlock) == 'B'
  lProtect := Eval(::protectBlock)
ENDIF

RETURN self

* --------------------

METHOD DC_XbpBrowse:Navigate( mp1, mp2 )
IF Valtype(::navigateBlock) == 'B'
  Eval(::navigateBlock)
ENDIF
IF mp1 == 12
  ::XbpBrowse:forceStable()
ENDIF
RETURN ::XbpBrowse:navigate( mp1, mp2 )

* --------------------

METHOD DC_XbpBrowse:forceStable()

LOCAL bErrorBlock

IF ::isTrapErrors
  bErrorBlock := ErrorBlock({|e|_TrapErrors(e)})
ENDIF

BEGIN SEQUENCE
::XbpBrowse:forceStable()
END SEQUENCE

IF ::isTrapErrors
  ErrorBlock(bErrorBlock)
ENDIF

RETURN self

* -----------

STATIC FUNCTION _TrapErrors()
BREAK
RETURN nil

* --------------------

METHOD DC_XbpBrowse:FieldBlock( nField )

RETURN {|x| iif(PCount()==0, ::dataSource:fieldGet(nField), ;
                             ::dataSource:fieldPut(nField, x))}

* ------------------

METHOD DC_XbpBrowse:save( cFileName, nStatus )

LOCAL aConfig[3], i, oColumn, nScale

nScale := ::getList:scaleFactor[3]
IF Empty(nScale)
  nScale := 1
ENDIF

nStatus := 1

aConfig[1] := ::currentPos()
aConfig[2] := ::currentSize()
aConfig[3] := {}

FOR i := 1 TO ::colCount
  oColumn := ::getColumn(i)
  Aadd( aConfig[3], { oColumn:columnPos, ;
                      { oColumn:currentSize()[1]/nScale, oColumn:currentSize()[2] }, ;
                      oColumn:origFonts[1], ;
                      oColumn:origFonts[2], ;
                      oColumn:heading:currentSize()[2], ;
                      oColumn:origFonts[3], ;
                      oColumn:footing:currentSize()[2] } )
NEXT

IF !Empty(cFileName)
  IF !DC_ASave(aConfig,cFileName)
    nStatus := 0
  ENDIF
ENDIF

RETURN aConfig

* ------------------

METHOD DC_XbpBrowse:restore( acFileName )

LOCAL aConfig, i, j, oColumn, aColumns[::colCount]

IF Valtype(acFileName) == 'C'
  aConfig := DC_ARestore(acFileName)
  IF Valtype(aConfig) # 'A'
    RETURN .f.
  ENDIF
ELSEIF Valtype(acFileName) == 'A'
  aConfig := acFileName
ELSE
  RETURN .f.
ENDIF

::lockUpdate(.t.)
FOR i := 1 TO ::colCount
  aColumns[i] := ::getColumn(i)
NEXT

FOR i := 1 TO Len(aColumns)
  ::delColumn(1)
NEXT
FOR i := 1 TO Len(aConfig[3])
  oColumn := NIL
  FOR j := 1 TO Len(aColumns)
    oColumn := aColumns[j]
    IF aConfig[3,i,1] == oColumn:columnPos
      EXIT
    ENDIF
  NEXT
  IF !oColumn==NIL
    ::addColumn(oColumn )
    oColumn:setSize(aConfig[3,i,2])
    oColumn:setHeaderCaption()
    oColumn:setHeaderColor()
    oColumn:setFooterCaption()
    oColumn:setFooterColor()
  ENDIF
NEXT
::setPos(aConfig[1])
::setSize(aConfig[2])
::lockUpdate(.f.)
::forceStable()
::xbpBrowse:refreshAll()

RETURN .t.

* -----------------

METHOD DC_XbpBrowse:CellEditSetup()

LOCAL aGetList, aGetListItem

aGetList := ::getList:getListArray
aGetListItem := ::getList:getListArray[::getListPointer]

DEFAULT aGetListItem[xGETLIST_OPTIONS2] := {}    // Cell Editing
//ASize(aGetListItem[xGETLIST_OPTIONS2],7)
ASize(aGetListItem[xGETLIST_OPTIONS2],8)                                      // J. Duijf 30-03-2013 Custom cell-edit keyblock
IF Valtype(aGetListItem[xGETLIST_OPTIONS2,1]) = 'N'
  aGetListItem[xGETLIST_OPTIONS2,1] := { aGetListItem[xGETLIST_OPTIONS2,1], nil }
ENDIF
DEFAULT aGetListItem[xGETLIST_OPTIONS2,1] := {0,nil}   // event
DEFAULT aGetListItem[xGETLIST_OPTIONS2,6] := ;
   ( aGetListItem[nGETLIST_SUBTYPE] == BROWSE_DATABASE )

IF aGetListItem[xGETLIST_OPTIONS2,1,1] > 0 // Edit Mode
  ::CellEditKeyBlock := aGetListItem[xGETLIST_OPTIONS2,8]                     // J. Duijf 30-03-2013
  ::cellEditMode := .t.
  ::getList:AddHotKey( aGetListItem[xGETLIST_OPTIONS2,1,1], ;
              DC_BrowCellBlock( ;
                self, ;
                aGetListItem[xGETLIST_OPTIONS2,3],;     // Edit mode
                aGetListItem[xGETLIST_OPTIONS2,7], ;    // Default value (array editing)
                aGetListItem[xGETLIST_OPTIONS2,2],;     // Edit Action block
                aGetListItem[xGETLIST_OPTIONS8], ;      // Handler block
                aGetListItem[xGETLIST_OPTIONS2,5],;     // Exit block
                aGetListItem[xGETLIST_OPTIONS9],;       // Reference Array
                aGetList),;
            nil, nil, self, self, nil, .t. )
ENDIF

DEFAULT aGetListItem[xGETLIST_OPTIONS3] := {}  // Cell Inserting
ASize(aGetListItem[xGETLIST_OPTIONS3],5)

DEFAULT aGetListItem[xGETLIST_OPTIONS3,1] := 0  // Insert Mode
IF aGetListItem[xGETLIST_OPTIONS3,1] > 0
  ::getList:AddHotKey( aGetListItem[xGETLIST_OPTIONS3,1], ;
               DC_BrowCellBlock( ;
                 self, ;
                 DCGUI_BROWSE_INSERT,;                   // Insert mode
                 aGetListItem[xGETLIST_OPTIONS3,3], ;    // Default value (array editing)
                 aGetListItem[xGETLIST_OPTIONS3,2], ;    // Action Block
                 aGetListItem[xGETLIST_OPTIONS8],   ;    // Handler Block
                 aGetListItem[xGETLIST_OPTIONS3,4], ;    // Exit Block
                 aGetListItem[xGETLIST_OPTIONS9],   ;    // Reference Array
                 aGetList), ;
            nil, nil, self, self, nil, .T.  )
  ::getList:AddHotkey( aGetListItem[xGETLIST_OPTIONS3,1], ;
                       {||SetAppFocus(self), ;
                          PostAppEvent(xbeP_Keyboard,aGetListItem[xGETLIST_OPTIONS3,1],,self)}, ;
                       {|o|o:=SetAppFocus(), ;
                           o:isDerivedFrom('XbpCellGroup') .AND. ;
                           o:parent:parent==self} )
ENDIF

DEFAULT aGetListItem[xGETLIST_OPTIONS3,5] := 0  // Append Mode
IF aGetListItem[xGETLIST_OPTIONS3,5] > 0
  ::getList:AddHotKey( aGetListItem[xGETLIST_OPTIONS3,5], ;
               DC_BrowCellBlock( ;
                 self, ;
                 DCGUI_BROWSE_APPEND,;                   // Append mode
                 aGetListItem[xGETLIST_OPTIONS3,3], ;    // Default value (array editing)
                 aGetListItem[xGETLIST_OPTIONS3,2], ;    // Action Block
                 aGetListItem[xGETLIST_OPTIONS8],   ;    // Handler Block
                 aGetListItem[xGETLIST_OPTIONS3,4], ;    // Exit Block
                 aGetListItem[xGETLIST_OPTIONS9],   ;    // Reference Array
                 aGetList ), ;
            nil, nil, self, self, nil, .T.  )

  ::getList:AddHotkey( aGetListItem[xGETLIST_OPTIONS3,5], ;
                       {||SetAppFocus(self), ;
                          PostAppEvent(xbeP_Keyboard,aGetListItem[xGETLIST_OPTIONS3,5],,self)}, ;
                       {|o|o:=SetAppFocus(), ;
                           o:isDerivedFrom('XbpCellGroup') .AND. ;
                           o:parent:parent==self} )
ENDIF

DEFAULT aGetListItem[xGETLIST_OPTIONS4] := {}  // Cell Deleting
ASize(aGetListItem[xGETLIST_OPTIONS4],3)
DEFAULT aGetListItem[xGETLIST_OPTIONS4,1] := 0

IF aGetListItem[xGETLIST_OPTIONS4,1] > 0  // Delete Mode
  ::getList:AddHotKey( aGetListItem[xGETLIST_OPTIONS4,1], ;
               DC_BrowCellBlock( ;
                 self, ;
                 DCGUI_BROWSE_DELETE,;                 // Delete mode
                 nil,;                                 // nil
                 aGetListItem[xGETLIST_OPTIONS4,2], ;  // Action block
                 aGetListItem[xGETLIST_OPTIONS8],;     // Handler block
                 aGetListItem[xGETLIST_OPTIONS4,3],;   // Exit block
                 aGetListItem[xGETLIST_OPTIONS9],;     // Reference Array
                 aGetList ), ;
            nil, nil, self, self, nil, .T. )

  ::getList:AddHotkey( aGetListItem[xGETLIST_OPTIONS4,1], ;
                       {||SetAppFocus(self), ;
                          PostAppEvent(xbeP_Keyboard,aGetListItem[xGETLIST_OPTIONS4,1],,self)}, ;
                       {|o|o:=SetAppFocus(), ;
                           o:isDerivedFrom('XbpCellGroup') .AND. ;
                           o:parent:parent==self} )
ENDIF

IF Valtype(aGetListItem[xGETLIST_OPTIONS2,1,2]) = 'B'

  ::keyBoard := {|a,b,o|DC_BrowCellEdit(self, ;
                        nil, ;
                        nil, ;
                        aGetListItem[xGETLIST_OPTIONS2,3], ;
                        nil, ;
                        aGetListItem[xGETLIST_OPTIONS2,7], ;
                        aGetListItem[xGETLIST_OPTIONS2,2], ;
                        self, ;
                        aGetListItem[xGETLIST_OPTIONS8], ;
                        aGetListItem[xGETLIST_OPTIONS2,5], ;
                        aGetListItem[xGETLIST_OPTIONS9], ;
                        aGetList, ;
                        .T., ;
                        Eval(aGetListItem[xGETLIST_OPTIONS2,1,2],a)) }
ENDIF

RETURN nil

* --------------------

METHOD DC_XbpBrowse:MoveColumn( nColumns, nPos, lRefresh )

LOCAL oColumn, i, lMoveLeft, bLeftButton, aSize, bRightButton, nNewPos, nSaveColorBG, ;
      bRightDataArea, aHeadingSize, aVisualStyle[3], aDrawMode[3], aCargo

DEFAULT nColumns := 0, ;
        lRefresh := .t., ;
        nPos := ::colPos

lMoveLeft := (nPos<::colCount)
IF nColumns > 0 .AND. nColumns + nPos > ::colCount
  nColumns := ::colCount - nPos
ELSEIF nColumns < 0 .AND. -1*nColumns > nPos
  nColumns := -1*nPos
ENDIF
IF nColumns == 0 .OR. ( nPos=1 .AND. nColumns<0 ) .OR. ;
  ( nPos==::colCount .AND. nColumns > 0 )
  RETURN self
ENDIF

IF nPos == ::sortColumn
  ::sortColumn := nPos + nColumns
ENDIF

::lockUpdate(.t.)
oColumn := ::getColumn(nPos)
nSaveColorBG := oColumn:heading:setColorBG()
bLeftButton := oColumn:heading:lbDown  // <<<<<<< save left sort block
bRightButton := oColumn:heading:rbDown  // <<<<<<< save right sort block
bRightDataArea := oColumn:dataArea:rbDown // <<<<<<< save right dataarea block
aVisualStyle[1] := oColumn:heading:useVisualStyle
aVisualStyle[2] := oColumn:dataArea:useVisualStyle
aVisualStyle[3] := oColumn:footing:useVisualStyle
aDrawMode[1] := oColumn:heading:drawMode
aDrawMode[2] := oColumn:dataArea:drawMode
aDrawMode[3] := oColumn:footing:drawMode
aCargo := oColumn:cargo

aSize := oColumn:currentSize()
aHeadingSize := oColumn:heading:currentSize()
::delColumn(nPos)
::lockUpdate(.f.)
nNewPos := nPos+nColumns
::insColumn(nNewPos,oColumn)
oColumn:setHeaderCaption()
oColumn:setFooterCaption()

IF ::cursorMode == XBPBRW_CURSOR_CELL
  IF nColumns > 0
    FOR i := 1 TO nColumns
      ::right()
    NEXT
  ELSE
    nColumns := -1 * nColumns
    IF lMoveLeft
      FOR i := 1 TO nColumns
        ::left()
      NEXT
    ENDIF
  ENDIF
ENDIF

IF !Empty(bLeftButton) .AND. 'oColumn:Sort' $ DC_GetCodeBlockText(bLeftButton)
   oColumn:heading:lbDown := _HeaderDown( self, oColumn, nNewPos )
ENDIF
IF !Empty(bRightButton) .AND. 'oColumn:Sort' $ DC_GetCodeBlockText(bRightButton)
   oColumn:heading:rbDown := _HeaderDown( self, oColumn, nNewPos )
ENDIF
IF !Empty(bRightDataArea)
   oColumn:dataArea:rbDown := bRightDataArea
ENDIF

IF Valtype(nSaveColorBG) == 'N'
  oColumn:heading:setColorBG(nSaveColorBG)
ENDIF
oColumn:setSize(aSize)
oColumn:heading:setSize(aHeadingSize)
oColumn := ::getColumn(nPos)

IF !Empty(oColumn:heading:lbDown) .AND. 'oColumn:Sort' $ DC_GetCodeBlockText(oColumn:heading:lbDown)
   oColumn:heading:lbDown := _HeaderDown( self, oColumn, nPos )
ENDIF
IF !Empty(oColumn:heading:rbdown) .AND. 'oColumn:Sort' $ DC_GetCodeBlockText(oColumn:heading:rbDown)
   oColumn:heading:rbDown := _HeaderDown( self, oColumn, nPos )
ENDIF
oColumn:heading:useVisualStyle := aVisualStyle[1]
oColumn:dataArea:useVisualStyle := aVisualStyle[2]
oColumn:footing:useVisualStyle := aVisualStyle[3]
oColumn:heading:drawMode := aDrawMode[1]
oColumn:dataArea:drawMode := aDrawMode[2]
oColumn:footing:drawMode := aDrawMode[3]
oColumn:cargo := aCargo

::setSize(::currentSize())
::forceStable()

IF lRefresh
  ::xbpBrowse:refreshAll()
ENDIF

RETURN self

* --------------------

METHOD DC_XbpBrowse:itemRbDown( aPos, aValues )

LOCAL nRow := aValues[2]
LOCAL nCol

IF ::rightButtonSelect

  #if XPPVER > 02000000
    nCol := aValues[1]:columnPos
  #else
    nCol := ::ColNum ( aValues[1] )
  #endif
  ::xbpBrowse:gotoItem( nRow, nCol )
  ::colPos := nCol
ENDIF

::XbpBrowse:itemRbDown( aPos, aValues )

RETURN self

* --------------------

METHOD DC_Xbpbrowse:cellEdit()

RETURN self

* --------------------

METHOD DC_XbpBrowse:cellEditMode(lMode)

Grok(lMode)

::cellEditMode := !::cellEditMode

RETURN self

* ----------------------

METHOD DC_XbpBrowse:SortClear()

LOCAL i, j, oColumn, oXbp, aChildList

FOR i := 1 TO ::colCount
  oColumn := ::GetColumn(i)
  IF Valtype(::sortColorNonselectableFG) = 'N'
    oColumn:heading:setColorFG(::sortColorNonselectableFG)
  ENDIF
  IF Valtype(::sortColorNonselectableBG) = 'N'
    oColumn:heading:setColorBG(::sortColorNonselectableBG)
  ENDIF
  IF oColumn:heading:status() > 0 .AND. !Empty(oColumn:heading:childList())
    aChildList := oColumn:heading:childList()
    FOR j := 1 TO Len(aChildList)
      oXbp := aChildList[j]
      IF oXbp:isDerivedFrom('XbpStatic') .AND. Valtype(oXbp:cargo) = 'C' .AND. ;
         oXbp:cargo == 'SortArrow'
        // destroy the arrow
        oXbp:destroy()
      ENDIF
    NEXT
  ENDIF
NEXT

RETURN self

* -----------------------

METHOD DC_XbpBrowse:HFLbDown ( aPos, aValues )

LOCAL i, aColumns[::colCount], nCol
FOR i := 1 TO Len(aColumns)
  aColumns[i] := ::getColumn(i)
  IF aColumns[i] == aValues[1]
    nCol := i
  ENDIF
NEXT

IF nCol > 1 .AND. aColumns[nCol-1]:noResize .AND. aPos[1] < 5
  RETURN nil
ELSEIF aColumns[nCol]:noResize .AND. aPos[1] > 5
  RETURN nil
ELSEIF aColumns[nCol]:noResize .AND. nCol < Len(aColumns) .AND. !aColumns[nCol+1]:noResize
  RETURN nil
ENDIF

RETURN ::XbpBrowse:HFLbDown( aPos, aValues )


METHOD DC_XbpBrowse:Tag( oCell, mp1, nKey )

LOCAL nCurrPos, nSaveRec, i, nKeyNo, nElement, nOffset, nSeconds := Seconds(), ;
      lUpdate := .t.

DEFAULT nKey := 0

IF nKey > 0 .AND. !(nKey $ { xbeK_CTRL_DOWN, xbeK_CTRL_UP, xbeK_SPACE, xbeK_CTRL_INS, xbeK_CTRL_DEL, xbeK_SH_DOWN })
  RETURN nil
ENDIF

IF nKey == xbeK_CTRL_DEL
  IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
    FOR i := 1 TO Len(::dataSource)
      ::dataSource[i,::tagElement] := .f.
    NEXT
  ELSEIF Valtype(::dataSource) == 'C'
    (::dataSource)->(DC_RecTagClear())
  ENDIF
  ::refreshAll()
ELSEIF nKey == xbeK_CTRL_INS
  IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
    FOR i := 1 TO Len(::dataSource)
      ::dataSource[i,::tagElement] := .t.
    NEXT
  ELSEIF Valtype(::dataSource) == 'C'
    nSaveRec := (::dataSource)->(RecNo())
    (::dataSource)->(DC_RecTagClear())
    (::dataSource)->(DC_DbGoTop())
    DO WHILE !(::dataSource)->(DC_Eof())
      (::dataSource)->(DC_RecTagToggle())
      (::dataSource)->(DC_DbSkip())
    ENDDO
    (::dataSource)->(dbGoTo(nSaveRec))
  ENDIF
  ::refreshAll()
ELSEIF AppKeyState(xbeK_CTRL) == 1 .OR. nKey $ { xbeK_CTRL_DOWN, xbeK_CTRL_UP, xbeK_SPACE }
  IF Valtype(oCell) == 'O'
    IF ::isLeftMouseButton
      nOffset := oCell:CellFromPos(mp1) - ::RowPos
      nElement := ::arrayElement + nOffset
      IF Valtype(::dataSource) == 'A' .AND. nElement > 0 .AND. nElement <= Len(::dataSource) .AND. ;
        Len(::dataSource[nElement]) >= ::tagElement .AND. Valtype(::dataSource[nElement,::tagElement]) == 'L'
        IF !::dataSource[nElement,::tagElement]
          ::dataSource[nElement,::tagElement] := .t.
          ::lastTagRec := nElement
          ::refreshAll()
        ENDIF
      // ELSEIF Valtype(::dataSource) == 'C'
      //   (::dataSource)->(DC_RecTagToggle())
      //   ::LastTagRec := (::dataSource)->(RecNo())
      ENDIF
    ENDIF
    RETURN nil
  ELSE
    IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
      ::dataSource[::arrayElement,::tagElement] := ;
         !::dataSource[::arrayElement,::tagElement]
      ::lastTagRec := ::arrayElement
    ELSEIF Valtype(::dataSource) == 'C'
      (::dataSource)->(DC_RecTagToggle())
      ::LastTagRec := (::dataSource)->(RecNo())
    ENDIF
    ::refreshCurrent()
    IF nKey == xbeK_CTRL_DOWN
      PostAppEvent( xbeP_Keyboard, xbeK_DOWN, nil, self )
    ELSEIF nKey == xbeK_CTRL_UP
      PostAppEvent( xbeP_Keyboard, xbeK_UP, nil, self )
    ENDIF
  ENDIF
  RETURN nil
ELSEIF AppKeyState(xbeK_SHIFT) == 1
  IF ::lastTagRec == 0
    IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
      ::dataSource[::arrayElement,::tagElement] := ;
         !::dataSource[::arrayElement,::tagElement]
    ELSEIF Valtype(::dataSource) == 'C'
      (::dataSource)->(DC_RecTagToggle())
    ENDIF
    ::refreshCurrent()
  ELSE
    IF Valtype(::dataSource) == 'A' .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
      nCurrPos := ::arrayElement
      FOR i := 1 TO Len(::dataSource)
        ::dataSource[i,::tagElement] := .f.
      NEXT
      IF nCurrPos > ::LastTagRec
        FOR i := ::LastTagRec TO nCurrPos
           ::dataSource[i,::tagElement] := .t.
        NEXT
      ELSEIF nCurrPos < ::LastTagRec
        FOR i := nCurrPos TO ::LastTagRec
           ::dataSource[i,::tagElement] := .t.
        NEXT
      ENDIF
    ELSEIF Valtype(::dataSource) == 'C'
      nSaveRec := (::dataSource)->(RecNo())
      nCurrPos := (::dataSource)->(DC_KeyNo())
      (::dataSource)->(DC_RecTagClear())
      IF nCurrPos > ::LastTagPos
        (::dataSource)->(DC_dbGoto(::LastTagRec))
        IF ::tagMode == DCGUI_TAGMODE_SET
          (::dataSource)->(DC_DbSkip())
        ENDIF
        DO WHILE (::dataSource)->(DC_KeyNo()) <= nCurrPos .AND. !(::dataSource)->(DC_Eof())
          (::dataSource)->(DC_RecTagToggle())
          (::dataSource)->(DC_DbSkip())
        ENDDO
      ELSEIF nCurrPos < ::LastTagPos
        (::dataSource)->(DC_dbGoto(::LastTagRec))
        IF ::tagMode == DCGUI_TAGMODE_SET
          (::dataSource)->(DC_DbSkip(-1))
        ENDIF
        DO WHILE (nKeyNo := (::dataSource)->(DC_KeyNo())) >= nCurrPos .AND. !(::dataSource)->(DC_Bof())
          (::dataSource)->(DC_RecTagToggle())
          (::dataSource)->(DC_DbSkip(-1))
          IF nKeyNo <= 1
            EXIT
          ENDIF
        ENDDO
      ENDIF
      (::dataSource)->(DC_dbGoto(nSaveRec))
    ENDIF
    ::refreshAll()
  ENDIF
  RETURN nil
ELSEIF Empty(oCell)
  IF Valtype(::dataSource) == 'C' .AND. !Empty((::dataSource)->(DC_RecTagArray()))
    (::dataSource)->(DC_RecTagClear())
    IF ::tagMode == DCGUI_TAGMODE_SET
      (::dataSource)->(DC_RecTagToggle())
    ELSE
      lUpdate := .f.
    ENDIF
  ELSEIF Valtype(::dataSource) = 'A' .AND. !Empty(::dataSource)
    FOR i := 1 TO Len(::dataSource)
      ::dataSource[i,::tagElement] := .f.
    NEXT
    IF ::tagMode == DCGUI_TAGMODE_SET .AND. ::arrayElement > 0 .AND. ::arrayElement <= Len(::dataSource) .AND. ;
      Len(::dataSource[::arrayElement]) >= ::tagElement .AND. Valtype(::dataSource[::arrayElement,::tagElement]) == 'L'
      ::dataSource[::arrayElement,::tagElement] := .t.
    ENDIF
  ELSE
    RETURN nil
  ENDIF
  ::refreshAll()
ELSE
  RETURN nil

ENDIF

IF lUpdate
  IF Valtype(::dataSource) == 'C'
    ::LastTagRec := (::dataSource)->(RecNo())
    ::LastTagPos := (::dataSource)->(DC_KeyNo())
  ELSEIF Valtype(::dataSource) == 'A'
    ::LastTagRec := ::arrayElement
    ::LastTagPos := ::arrayElement
  ENDIF
ENDIF

RETURN nil

//
// J. Duijf  21-04-2010
// Zebra stripes utilisation
//
METHOD DC_XbpBrowse:ZebraColor(uEven,aOdd)
//
// J. Duijf 11-03-2011 Get default colors from presentation parameters
//
LOCAL aRet           := nil
//
do case
   case ValType(uEven) = "A"                                            // Array with foreground / background color
      if ::evenRow                                                      // Rows 2,4,6 etc
         aRet := uEven
      else                                                              // Rows 1,3,5 etc
         aRet := aOdd
      endif
   case ValType(uEven) = "B"                                            // User Codeblock
      aRet := Eval(uEven,::EvenRow)
   case ValType(aOdd) = "A"                                             // Array with foreground / background color
      if ::evenRow                                                      // Rows 2,4,6 etc
         aRet := aOdd
      endif
endcase

if ValType(aRet) <> "A"
   aRet              := {DC_GetPresentation(::browsePres,XBP_PP_COL_DA_FGCLR,GRA_CLR_BLACK),;
                         DC_GetPresentation(::browsePres,XBP_PP_COL_DA_BGCLR,GRA_CLR_WHITE)}
   //
   // J. Duijf 11-03-2011
   // Workaround for PDR 6365
   // The colorBlock should return a 4 element array to prevent an incorrect
   // Hilite forground color.
   // {CellFGColor,CellBGColor,CellHiliteFGColor,CellHiliteBGColor}
   // The highlite colors come from the presentation parameters.
   // If undefined, the default BG/FG hilite colors are BLUE/WHITE
   //
   Aadd(aRet,DC_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_FGCLR,GRA_CLR_WHITE))
   Aadd(aRet,DC_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_BGCLR,GRA_CLR_BLUE))
else
   if Len(aRet) = 2
      //
      // J. Duijf 11-03-2011
      // Workaround for PDR 6365
      // The colorBlock should return a 4 element array to prevent an incorrect
      // Hilite forground color.
      // {CellFGColor,CellBGColor,CellHiliteFGColor,CellHiliteBGColor}
      // The highlite colors come from the presentation parameters.
      // If undefined, the default BG/FG hilite colors are BLUE/WHITE
      //
      Aadd(aRet,DC_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_FGCLR,GRA_CLR_WHITE))
      Aadd(aRet,DC_GetPresentation(::browsePres,XBP_PP_COL_DA_HILITE_BGCLR,GRA_CLR_BLUE))
   endif
endif
Return aRet
* --------------------

STATIC FUNCTION _RowColor( oBrowse, aOptions, aColor, xData, bColorBlock )

LOCAL bRowColor

/*
 IF Valtype( bColorBlock ) = 'B' ; RETURN bColorBlock
 ELSEIF Empty(aOptions[1]) .AND. Empty(aOptions[3]) // MARK and MARKEVAL clauses not used
*/
//
// J. Duijf 28-04-2011
// bColorBlock can already be assigned (In case of Zebra stripes), but the MARK and MKEVAL clauses stil need to be evaluated.
//

if Empty(aOptions[1]) .AND. Empty(aOptions[3]) // MARK and MARKEVAL clauses not used
   IF Valtype( bColorBlock ) <> 'B'
      IF Valtype(DC_BrowseColor()) == 'A' .OR. oBrowse:tagEnable
        bRowColor := {||oBrowse:rowColor()}
      ELSE
        bRowColor := nil
      ENDIF
   else
      bRowColor   := bColorBlock
   ENDIF
ELSEIF Valtype(xData) = 'A'
/*
  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := oBrowse:mark, ;
                             g := IIF(a[n,f], ;
                                  { aOptions[5], aOptions[6] }, aColor ) }
*/
//
// Modified 28-04-2011 J. Duijf,
// If the marked color clause is used, and a row is NOT marked, then use bColorBlock if available.
//
  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := oBrowse:mark, ;
                             g := IIF(a[n,f], ;
                                  { aOptions[5], aOptions[6] }, iif(ValType(bColorBlock) = "B",Eval(bColorBlock),aColor)) }
ELSE
/*
  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := IIF(Valtype(oBrowse:markColorBlock)='B', ;
                                  oBrowse:markColorBlock,oBrowse:mark), ;
                             g := IIF(Eval(f), ;
                                  { aOptions[5], aOptions[6] }, aColor ) }
//
// Modified 3-1-2003, J. Duijf DCBROWSE does also support codeblocks as rowcolor
//                [MKCOLOR <nbMarkEval>, <ncMFgC> [,<ncMBgC>] ]               ;
//
// <ncMFgC> can now be a codeblock. It must return {FG,BG}
//
  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := IIF(Valtype(oBrowse:markColorBlock)='B', ;
                                  oBrowse:markColorBlock,oBrowse:mark), ;
                             g := IIF(Eval(f), ;
                                  iif(ValType(aOptions[5]) ='B',Eval(aOptions[5]), ;
                                  { aOptions[5], aOptions[6] }), aColor ) }

 Modified 7-03-2010 J. Duijf DCBROWSE the codeblock for color receives the browse
                             object as first parameter, this for utilisation of zebra patern

  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := IIF(Valtype(oBrowse:markColorBlock)='B', ;
                                  oBrowse:markColorBlock,oBrowse:mark), ;
                             g := IIF(Eval(f), ;
                                  iif(ValType(aOptions[5]) ='B',Eval(aOptions[5],oBrowse), ;
                                  { aOptions[5], aOptions[6] }), aColor ) }

*/
// Modified 28-04-2011 J. Duijf,
// If the MKCOLOR clause is used, and a row is NOT marked, then use bColorBlock if available.
//

  bRowColor :=  { |x,a,n,f,g | a := oBrowse:dataSource, ;
                             n := oBrowse:arrayElement, ;
                             f := IIF(Valtype(oBrowse:markColorBlock)='B', ;
                                  oBrowse:markColorBlock,oBrowse:mark), ;
                             g := IIF(Eval(f), ;
                                  iif(ValType(aOptions[5]) ='B',Eval(aOptions[5],oBrowse), ;
                                  { aOptions[5], aOptions[6] }), iif(ValType(bColorBlock) = "B",Eval(bColorBlock),aColor)) }

ENDIF

RETURN bRowColor


* ===============================================

CLASS DC_MultiLineCellGroup FROM XbpCellGroup

EXPORTED:
  VAR width
  VAR presentation
  VAR rows
  VAR noResize

  INLINE METHOD Create( oParent, oOwner, aPos, aSize, aPP, lVisible )

  LOCAL i

  IF Len(::presentation) > 0
    IF Valtype(::presentation[2]) = 'A'
      FOR i := 1 TO Len(::presentation[2])
        DC_PresAdd(aPP,::presentation[2,i])
      NEXT
    ELSEIF Valtype(::presentation[1]) = 'A'
      FOR i := 1 TO Len(::presentation[1])
        DC_PresAdd(aPP,::presentation[1,i])
      NEXT
    ENDIF
  ENDIF

  IF ::noResize
    aSize := { ::width, -1 }
  ENDIF
  ::maxRow := ::rows
  if ::MaxRow > 1
     aSize[2] := -1
  endif
  IF IsMemberVar(self,'USEVISUALSTYLE')
    ::xbpCellGroup:useVisualStyle := .f.
  ENDIF

  Return ::XbpCellgroup:create( oParent, oOwner, aPos, aSize, aPP, lVisible )

ENDCLASS

* ===================================================

CLASS DC_XbpColumn FROM DC_XbpColumnSuper, DC_XbaseParts
HIDDEN:
VAR      nOffset                             // J. Duijf 7-03-2010
VAR      LastRow                             // J. Duijf 7-03-2010
EXPORTED:
  VAR validBlock
  VAR sortBlock
  VAR cellEditGetList
  VAR cellEditExitKey
  VAR getsetBlock
  VAR headerBlock
  VAR headerCaption
  VAR headerFont
  VAR headerHeight
  VAR dataSource
  VAR parent
  VAR picture
  VAR cellEditID
  VAR footerBlock
  VAR footerCaption
  VAR footerFont
  VAR footerHeight
  VAR pointer
  VAR presArray
  VAR columnPos
  VAR arrayPointer
  VAR headerColorBlock
  VAR footerColorBlock
  VAR sortDefault
  VAR cellEditor
  VAR noResize
  VAR dataLinkMultiLine
  VAR columnSize
  VAR noAutoResizeColumn
  VAR validateAlways
  VAR origFonts
  VAR visualStyleArray
  VAR drawModeArray

  METHOD Init, setHeaderCaption, setFooterCaption, setHeaderColor, setFooterColor, ;
         MoveColumn, Refresh, WhenHideEval, Sort, resizeColumn, Destroy

  METHOD DrawRow, ScrollUp, ScrollDown       // J. Duijf 7-03-2010
ENDCLASS

* ---------------------

METHOD DC_XbpColumn:Destroy()
::DC_XbpColumnSuper:Destroy()
/*
::datasource      := nil
::parent          := nil
::Footing         := nil
::Heading         := nil
::Cargo           := nil
::datalink        := nil
::getlist         := nil
::xbp             := nil
*/
Return SELF

METHOD DC_XbpColumn:init( oParent, oGetList, lTest )

LOCAL aHeader, cCaption, xOptions, aArray, aGetList, aGetListItem, ;
      nPointer, aFooter, aColor, lPixel, bBlock, bDataBlock, ;
      bColorBlock, xFont, aPres, lMultiline, ;
      nColorBG, aPP, nFound, nColorHiliteFG, nWidth
LOCAL bColorTmpBlock
DEFAULT lTest := .f.
::visualStyleArray := Array(3)
::drawModeArray := Array(3)

::nOffset      := 0                          // J. Duijf 7-03-2010

::DC_XbaseParts:init( oGetList,,self )

aGetList := ::getList:getListArray
aGetListItem := ::getList:getListArray[::getListPointer]

DEFAULT aGetListItem[nGETLIST_SUBTYPE] := XBPCOL_TYPE_TEXT
DEFAULT aGetListItem[nGETLIST_WIDTH]   := XBP_AUTOSIZE

xOptions := aGetListItem[xGETLIST_OPTIONS]

aArray   := Aclone(oParent:browsePres)                                        // J. Duijf 10-03-2011
ASize(xOptions,13)
IF Valtype( _Pres(aGetListItem[aGETLIST_PRESENTATION]) ) = 'A'
   //
   // J. Duijf 10-03-2011
   // Instead of replacing the array, properties are merged.
   // aArray := AClone(_Pres(aGetListItem[aGETLIST_PRESENTATION]))
   //
   Dc_MergePresentation(aArray,_Pres(aGetListItem[aGETLIST_PRESENTATION]))
//ELSE
//  aArray := {}
//
ENDIF

DEFAULT xOptions[5] := xbeK_ENTER
DEFAULT xOptions[6] := .f.  // left button
DEFAULT xOptions[7] := .f.  // no create
DEFAULT xOptions[9] := .f.  // no resize
DEFAULT xOptions[11] := .f. // no resizecolumns
DEFAULT xOptions[12] := .f. // valid always
DEFAULT xOptions[13] := .f. // owner draw

::noResize := xOptions[9]
::noAutoResizeColumn := xOptions[11]
::validateAlways := xOptions[12]

// get pointer to cell editor
::cellEditID := xOptions[4]
IF Valtype(::cellEditID) = 'C' .AND. !Empty(::cellEditID)  // Custom Editor ID
  nPointer := AScan(aGetList,{|a|a[cGETLIST_ID]==::cellEditID})
  IF nPointer > 0
    ::cellEditGetList := { aGetList[nPointer] } // , aGetList[nDataStore] }
  ENDIF
ENDIF

// get pointer to array element
nPointer := aGetListItem[nGETLIST_POINTER]
::arrayPointer := nPointer

IF Valtype( nPointer ) = 'N' // it's an array element pointer
  bDataBlock := DC_GetColBlock( nPointer, oParent )
  aGetListItem[bGETLIST_VAR] := bDataBlock
ELSE // it's a column code block
  bDataBlock := DC_GetBlock( aGetListItem[bGETLIST_VAR] )
  IF Valtype(bDataBlock) # 'B'
    bDataBlock := DC_GetBlock( aGetListItem[bGETLIST_REFVAR] )
  ENDIF
ENDIF

cCaption := aGetListItem[cGETLIST_CAPTION]
IF Valtype(cCaption) = 'B'
  ::headerBlock := cCaption
  // cCaption := Eval(cCaption)
  cCaption := Chr(255)
ENDIF

IF Valtype(cCaption) = 'C'
  aHeader := DC_TokenArray( cCaption, oParent:headerDelim )
  ::headerCaption := cCaption
ELSEIF Valtype(cCaption) = 'A'
  aHeader := cCaption
ELSEIF oParent:hasHeader
  aHeader := {Chr(255)}
ELSE
  aHeader := {nil}
ENDIF

ASize(aHeader,oParent:headerLines)

/*
::resize := {|a,b,o,aChildren|aChildren := o:heading:childList(), ;
              IIF(!Empty(aChildren),;
              aChildren[1]:setPos({o:currentSize()[1]-20,aChildren[1]:currentPos()[2]}),nil)}
*/
//
// J. Duijf 19-02-2012
// If there is more then 1 header line, ignore the fixed header height
// It must be recalculated based on the fontsize (dynamic)
//
if Len(aHeader) > 0
   Dc_MergePresentation(aArray,{{XBP_PP_COL_HA_HEIGHT,-1}})
endif
//

cCaption := xOptions[3]

IF Valtype(cCaption) = 'B'
  cCaption := Eval(cCaption)
ENDIF

IF Valtype(cCaption) = 'C'
  aFooter := DC_TokenArray( cCaption, oParent:footerDelim )
  ::footerCaption := cCaption
//
// J. Duijf 18-09-2013
// Also blank footer line possible
//
  If Len(aFooter) = 0
     aFooter := {cCaption}
  Endif
ELSEIF Valtype(cCaption) = 'A'
  aFooter := cCaption
ELSEIF oParent:hasFooter
  aFooter := {Chr(255)}
ELSE
  aFooter := {nil}
ENDIF

ASize(aFooter,oParent:footerLines)
//
// J. Duijf 19-02-2012
// If there is more then 1 footer line, ignore the fixed footer height
// It must be recalculated based on the fontsize (dynamic)
//
if Len(aFooter) > 0
   Dc_MergePresentation(aArray,{{XBP_PP_COL_FA_HEIGHT,-1}})
endif
//
IF !( aHeader[1] == NIL )
  DC_PresAdd( aArray, { XBP_PP_COL_HA_CAPTION, aHeader[1] } )
  oParent:hasHeader := .t.
ENDIF

lPixel := aGetListItem[lGETLIST_PIXEL]

nWidth := aGetListItem[nGETLIST_WIDTH]

IF lPixel
  nWidth *= ::getList:scaleFactor[3]
  DC_PresAdd( aArray, { XBP_PP_COL_DA_ROWWIDTH, nWidth } )
  DC_PresAdd( aArray, { XBP_PP_COL_DA_CHARWIDTH, nWidth } )
ELSE
  DC_PresAdd( aArray, { XBP_PP_COL_DA_CHARWIDTH, nWidth } )
ENDIF

IF !( aFooter[1] == NIL )
  DC_PresAdd( aArray, { XBP_PP_COL_FA_CAPTION, aFooter[1] } )
  oParent:hasFooter := .t.
ENDIF
//
// J. Duijf 12-03-2011 Optional allignment has prefference
//

::parent := oParent

IF Valtype(aGetListItem[xGETLIST_OPTIONS7]) = 'N'

  if Ascan( aArray, {|a|a[1] = XBP_PP_COL_DA_CELLALIGNMENT}) = 0                  // No default
    DC_PresAdd( aArray, { XBP_PP_COL_DA_CELLALIGNMENT, aGetListItem[xGETLIST_OPTIONS7] })
  Endif

  // J. Duijf 20-06-2012 Thanks to Clifford Wiernick,
  // Only change if alignment is NOT set by previous settings
  //
  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_HA_ALIGNMENT}) = 0                  // No default
     DC_PresAdd( aArray, { XBP_PP_COL_HA_ALIGNMENT,  aGetListItem[xGETLIST_OPTIONS7] } )
  Endif
  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_FA_ALIGNMENT}) = 0                  // No default
     DC_PresAdd( aArray, { XBP_PP_COL_FA_ALIGNMENT,  aGetListItem[xGETLIST_OPTIONS7] } )
  Endif

ELSEIF Valtype(bDataBlock) = 'B' .AND. Valtype(Eval(bDataBlock,::parent:recordObject)) = 'N'        // Default handling

  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_DA_CELLALIGNMENT}) = 0                  // No default
    DC_PresAdd( aArray, { XBP_PP_COL_DA_CELLALIGNMENT, XBPALIGN_RIGHT + XBPALIGN_VCENTER } )
  ENDIF
  //
  // J. Duijf 10-03-2011
  // If a data item is numeric, and Right alligned, so should be the header and footer
  //
  // J. Duijf 12-03-2011 Not sure why i do not need to add XBPALIGN_VCENTER
  //                     The header and footer are centered anyway (and do not if XBPALIGN_VCENTER is added)
  //
  // J. Duijf 20-06-2012 Thanks to Clifford Wiernick,
  // Only change if alignment is NOT set by previous settings
  //
  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_HA_ALIGNMENT}) = 0                  // No default
     DC_PresAdd( aArray, { XBP_PP_COL_HA_ALIGNMENT, XBPALIGN_RIGHT}) // + XBPALIGN_VCENTER } )
  Endif
  If Ascan( aArray, {|a|a[1] = XBP_PP_COL_FA_ALIGNMENT}) = 0                  // No default
     DC_PresAdd( aArray, { XBP_PP_COL_FA_ALIGNMENT, XBPALIGN_RIGHT}) // + XBPALIGN_VCENTER } )
  Endif

ENDIF

xFont := aGetListItem[cGETLIST_FONT]
IF !Empty(xFont)
  IF Valtype(xFont) == 'B'
    xFont := Eval(xFont)
  ENDIF
  DC_PresAdd( aArray, { XBP_PP_COMPOUNDNAME, xFont } )
ENDIF

::XbpColumn:init(,,,,aArray, .t. )

IF xOptions[13]
  ::drawMode := XBP_DRAW_OWNER
ENDIF

IF !Empty(aGetListItem[nGETLIST_SUBTYPE])
  ::type := aGetListItem[nGETLIST_SUBTYPE]
ENDIF
::picture := aGetListItem[cGETLIST_PICTURE]

aPres := {oParent:headerPres,aGetListItem[xGETLIST_OPTIONS3]}
::Heading := DC_MultilineCellgroup():New( self,,,,,,Len(aHeader) )
::Heading:width := aGetListItem[nGETLIST_WIDTH]
::Heading:presentation := aPres
::Heading:rows := Max(Len(aHeader),oParent:Headerlines)
::Heading:noResize := ::noResize
::Heading:ReferenceString  := "W"

aPres := {oParent:footerPres,aGetListItem[xGETLIST_OPTIONS4]}
::Footing := DC_MultilineCellgroup():New( self,,,,,,Len(aFooter) )
::Footing:width := aGetListItem[nGETLIST_WIDTH]
::Footing:presentation := aPres
::Footing:rows := Max(Len(aFooter),oParent:FooterLines)
::Footing:noResize := ::noResize
::Footing:ReferenceString  := "W"

aColor := aGetListItem[aGETLIST_COLOR]

::sortBlock := DC_GetBlock( xOptions[1] )
::sortDefault := xOptions[8]

IF Valtype(::sortDefault) == 'B'
  ::sortDefault := Eval(::sortDefault,self)
  DEFAULT ::sortDefault := .f.
ENDIF

::hideBlock := DC_GetBlock(aGetListItem[bGETLIST_HIDE])
::validBlock := DC_GetBlock(aGetListItem[bGETLIST_VALID])
::whenBlock := DC_GetBlock(aGetListItem[bGETLIST_WHEN])

::cargo := { ::getListPointer, aGetList, aGetListItem[xGETLIST_CARGO] }

::cellEditExitKey := xOptions[5]
::getsetBlock := {|x|Eval(bDataBlock,x,self)}
::headerBlock := aGetListItem[cGETLIST_CAPTION]
::dataSource := oParent:dataSource
::protectBlock := DC_GetBlock(aGetListItem[bGETLIST_PROTECT])
::footerBlock := xOptions[3]
::pointer := nPointer
::presArray := aArray
IF Valtype(aGetListItem[cGETLIST_LEVEL]) = 'A'
  ::columnPos := aGetListItem[cGETLIST_LEVEL,2]
ELSE
  ::columnPos := oParent:colCount + 1
ENDIF

bColorBlock := _ColorBlock( aColor )

/*
if ValType(bColorBlock) = "B"
   //
   // J. Duijf 11-03-2011
   //
   // Workaround for PDR 6365
   // The colorBlock should return nil, or a 4 element array to prevent an incorrect
   // Hilite forground color.
   // {CellFGColor,CellBGColor,CellHiliteFGColor,CellHiliteBGColor}
   // The highlite colors come from the presentation parameters.
   // If undefined, the default BG/FG hilite colors are BLUE/WHITE
   // If the colorblock returns NIL, the colors come from oParent:rowColorBlock
   //
   bColorTmpBlock    := bColorBlock

   bColorBlock       := {|a|a := Eval(bColorTmpBlock,Eval(::dataLink)),;
                              iif(ValType(a) = "A" .and. Len(a) = 2,;
                                 (Aadd(a,Dc_GetPresentation(::presArray,XBP_PP_COL_DA_HILITE_FGCLR,GRA_CLR_WHITE)), ;
                                  Aadd(a,Dc_GetPresentation(::presArray,XBP_PP_COL_DA_HILITE_BGCLR,GRA_CLR_BLUE))),;
                                  (a := iif(ValType(a) = "A",a,;
                                   iif(ValType(oParent:rowColorBlock) = "B",Eval(oParent:rowColorBlock,oParent),nil)))),a}
endif
*/

IF Valtype(bColorBlock) = 'B'
  ::colorBlock := bColorBlock
ELSEIF Valtype(oParent:rowColorBlock) = 'B'                 // Color of parent (Dc_Browse)
//  ::colorBlock := oParent:rowColorBlock
  ::colorBlock := {||Eval(oParent:rowColorBlock,oParent)}   // J. Duijf 7-03-2010
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_PREEVAL])
IF Valtype(bBlock) = 'B'
  Eval( bBlock, self )
ENDIF

lMultiline := (Valtype(bDataBlock) = 'A')

IF !Empty(::picture)
  ::dataLink := _ColPicture( bDataBlock, ::picture, self, ::parent:recordObject )
ELSEIF !lMultiline .AND. Valtype(::parent:dataSource) # 'O'
  IF Valtype(::parent:recordObject) == 'O'
    ::dataLink := {||Eval(bDataBlock,::parent:recordObject,self)}
  ELSE
    ::dataLink := {|x|Eval(bDataBlock,x,self)}
  ENDIF
ELSE
  ::dataLink := bDataBlock
ENDIF

IF lMultiLine
  ::dataArea:= DC_XbpCellGroupMultiLine():New( self )
  ::dataLinkMultiLine := ::dataLink
  ::dataLink := ::dataLinkMultiLine[1]
ENDIF

// ::XbpColumn:create( oParent,,,,aArray,.t. )
::presArray := aArray

oParent:addColumn( self )

IF lMultiline

  nColorHiliteFG := XBPSYSCLR_HILITEBACKGROUND
  aPP := oParent:setPresParam()
  nFound := AScan(aPP,{|a|a[1]==XBP_PP_COL_DA_HILITE_BGCLR})
  IF nFound = 0
    nFound := AScan(aPP,{|a|a[1]==XBP_PP_HILITE_BGCLR})
  ENDIF
  IF nFound > 0
    nColorHiliteFG := aPP[nFound,2]
  ENDIF
  nColorBG := ::dataArea:setColorBG()
  AADD( aArray, { XBP_PP_COL_DA_HILITE_FGCLR, nColorHiliteFG } )

  AADD( aArray, { XBP_PP_COL_DA_FGCLR, nColorBG } )

  ::XbpColumn:configure(,,,,aArray,.t.)
  ::dataLink := {||{}}

  ::dataArea:dataLinkMultiLine := ::dataLinkMultiLine
  ::dataArea:column            := self
  ::dataArea:browse            := ::setParent()

  DO WHILE !::dataArea:browse:isDerivedFrom( "XbpBrowse" )
    ::dataArea:browse := ::dataArea:browse:setParent()
  ENDDO

ENDIF

IF Valtype(::hideBlock)='B' .AND. Eval(::hideBlock)
  ::hide()
ENDIF

IF oParent:optimize // optimize
  ::heading:lbDblClick := {|a,b,o|_Optimize(a,b,o)}
ENDIF

::dataArea:keyboard := oParent:keyboard

// ::dataArea:lbDown := {|a,b,o|PostAppEvent(xbeP_SetInputFocus,,,self)}
IF oParent:tagEnable
  ::dataarea:LbClick := DC_MergeBlocks(::dataArea:LbClick,{||oParent:Tag()} )
  ::dataArea:Motion := {|mp1,b,oCell|oParent:Tag(oCell,mp1,0)}
  ::dataArea:lbDown := DC_MergeBlocks(::dataArea:lbDown,{||oParent:isLeftMouseButton := .t.})
  ::dataArea:lbUp := {||oParent:isLeftMouseButton := .f.}
ENDIF

::Heading:cargo := aGetListItem[cGETLIST_CAPTION]
::setHeaderCaption(::heading:cargo)

::Footing:cargo := xOptions[3]

::setFooterCaption(::footing:cargo)

IF Valtype(oParent:sortColorUnselectedBG)=='N' .AND. Valtype(::sortBlock)=='B'
  ::Heading:setColorBG(oParent:sortColorUnselectedBG)
ELSEIF Valtype(oParent:sortColorNonselectableBG)=='N'
  ::Heading:setColorBG(oParent:sortColorNonselectableBG)
ENDIF

IF Valtype(oParent:sortColorUnselectedFG)=='N' .AND. Valtype(::sortBlock)=='B'
  ::Heading:setColorFG(oParent:sortColorUnselectedFG)
ELSEIF Valtype(oParent:sortColorNonselectableFG)=='N'
  ::Heading:setColorFG(oParent:sortColorNonselectableFG)
ENDIF

aColor := aGetListItem[xGETLIST_OPTIONS2]
IF Valtype(aColor) = 'A' // header color
  ASize(aColor,4)
  IF Valtype(aColor[1])='B'
    aColor := Eval(aColor[1])
    ASize(aColor,4)
  ENDIF
  IF Valtype(aColor[1]) = 'A'
    aColor[1] := GraMakeRGBColor(aColor[1])
  ENDIF
  IF Valtype(aColor[2]) = 'A'
    aColor[2] := GraMakeRGBColor(aColor[2])
  ENDIF
  IF Valtype(aColor[1])='N'
    ::heading:setColorFG(aColor[1])
  ENDIF
  IF Valtype(aColor[2])='N'
    ::heading:setColorBG(aColor[2])
  ENDIF
  ::headerFont := aColor[3]
  ::headerHeight := aColor[4]
ENDIF

aColor := aGetListItem[xGETLIST_OPTIONS8]
IF Valtype(aColor) = 'A' // footer color
  ASize(aColor,4)
  IF Valtype(aColor[1])='B'
    aColor := Eval(aColor[1])
    ASize(aColor,4)
  ENDIF
  IF Valtype(aColor[1]) = 'A'
    aColor[1] := GraMakeRGBColor(aColor[1])
  ENDIF
  IF Valtype(aColor[2]) = 'A'
    aColor[2] := GraMakeRGBColor(aColor[2])
  ENDIF
  IF Valtype(aColor[1])='N'
//    DC_PresAdd( aArray, { XBP_PP_COL_FA_FGCLR, aColor[1] } )
  ENDIF
  IF Valtype(aColor[2])='N'
//    DC_PresAdd( aArray, { XBP_PP_COL_FA_BGCLR, aColor[2] } )
  ENDIF
  IF Valtype(aColor[3]) == 'C'
    ::footerFont := aColor[3]
  ENDIF
  IF Valtype(aColor[4]) == 'N'
    ::footerHeight := aColor[4]
  ENDIF
ENDIF

IF Valtype(aGetListItem[xGETLIST_OPTIONS2]) = 'A' .AND. ;
   Valtype(aGetListItem[xGETLIST_OPTIONS2,1]) = 'B'
  ::headerColorBlock := aGetListItem[xGETLIST_OPTIONS2,1]
  ::setHeaderColor()
ENDIF

IF Valtype(aGetListItem[xGETLIST_OPTIONS8]) = 'A' .AND. ;
   Valtype(aGetListItem[xGETLIST_OPTIONS8,1]) = 'B'
  ::footerColorBlock := aGetListItem[xGETLIST_OPTIONS8,1]
  ::setFooterColor()
ENDIF

IF !Empty(::headerFont)
  ::heading:setFontCompoundName(::headerFont)
ENDIF
IF !Empty(::headerHeight)
  ::heading:setSize({::heading:currentSize()[1],::headerHeight})
ENDIF

IF !Empty(::footerFont)
  ::footing:setFontCompoundName(::footerFont)
ENDIF
IF !Empty(::footerHeight)
  ::footing:setSize({::footing:currentSize()[1],::footerHeight})
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_GROUP],'O')
IF Valtype( bBlock ) = 'B'
  Eval( bBlock, self )
ENDIF

bBlock := DC_GetBlock(aGetListItem[bGETLIST_GOTFOCUS])
IF Valtype( bBlock ) = 'B'
  // ::setInputFocus := bBlock
  ::dataArea:setInputFocus := bBlock
ENDIF
bBlock := DC_GetBlock(aGetListItem[bGETLIST_LOSTFOCUS])
IF Valtype( bBlock ) = 'B'
  ::killInputFocus := bBlock
  ::dataArea:killInputFocus := bBlock
ENDIF

IF Valtype(oParent:rightButtonDown) = 'B'
  ::dataArea:RbDown  := DC_MergeBlocks(::dataArea:RbDown,oParent:rightButtonDown)
ENDIF

IF Valtype(::sortBlock) = 'B'
  IF xOptions[6]
    ::heading:lbdown := _HeaderDown( oParent, self )
  ENDIF
  ::heading:rbdown := _HeaderDown( oParent, self )
ELSEIF ::noResize
ENDIF

DEFAULT ::sortDefault := .t.
IF ::sortDefault
  ::sort( nil, oParent:colCount )
ENDIF

::getList:isDataToolTips := ::getList:isDataToolTips .OR. Valtype(aGetListItem[xGETLIST_OPTIONS9]) == 'A'
DEFAULT aGetListItem[xGETLIST_OPTIONS9] :=  {{||.f.},nil,'',0,0,0,nil,nil}
ASize(aGetListItem[xGETLIST_OPTIONS9],8)
aGetListItem[xGETLIST_OPTIONS9,7] := oParent // pointer to browse object
aGetListItem[xGETLIST_OPTIONS9,8] := oParent:dataSource
::dataarea:cargo := aGetListItem[xGETLIST_OPTIONS9]  // DataArea tooltip enable codeblock

// :cargo[1] = Enable Code Block
// :cargo[2] = Return Code Block
// :cargo[3] = Value to display in tooltip
// :cargo[4] = Last moused cell record offset
// :cargo[5] = Last selected record
// :cargo[6] = Last moused cell record
// :cargo[7] = Pointer to Browse object
// :cargo[8] = Alias(), Array or SQLExpress object

::DC_XbaseParts:setOrig(self)

IF IsMemberVar(oParent,'USEVISUALSTYLE')
  ::heading:useVisualStyle := oParent:useVisualStyle
  ::footing:useVisualStyle := oParent:useVisualStyle
ENDIF

::origFonts := ;
   { ::dataArea:setFontCompoundName(), ;
     ::heading:setFontCompoundName(), ;
     ::footing:setFontCompoundName() }

::visualStyleArray[1] := ::heading:useVisualStyle
::visualStyleArray[2] := ::dataArea:useVisualStyle
::visualStyleArray[3] := ::footing:useVisualStyle
::drawModeArray[1] := ::heading:drawMode
::drawModeArray[2] := ::dataArea:drawMode
::drawModeArray[3] := ::footing:drawMode
RETURN self

* --------------------

METHOD DC_XbpColumn:resizeColumn( aOldSize, aNewSize )

::setSize({::currentSize()[1]+(aNewSize[1]-aOldSize[1]),::CurrentSize()[2]})
::parent:setSize(::parent:currentSize())

RETURN self

*----------------------

METHOD DC_XbpColumn:refresh()

::setHeaderCaption()
::setHeaderColor()
::setFooterCaption()
::setFooterColor()

RETURN self

* --------- This is from build 227. It works for John Lau,
* --------- The commented out version was from 232.
* --------- It didn't work for him.  I don't know why.
* --------- I'd like to find a version that works for everyone
* --------- For now, wer'e going with this until somebody complains

METHOD DC_XbpColumn:WhenHideEval( lEnable, lShow, lProtect )

LOCAL oParent, nColumn, j, oColumn, nInsColumn, lOk

Grok(lShow)

IF Valtype(::hideBlock) = 'B'
  oParent := ::parent
  nColumn := ::columnPos
  IF Eval(::hideBlock,self)
    lOk := .f.
    FOR j := 1 TO oParent:colCount
      oColumn := oParent:getColumn(j)
      IF oColumn:columnPos == nColumn // column has not been deleted
        lOk := .t.
        EXIT
      ENDIF
    NEXT
    IF lOk
      oParent:lockUpdate(.t.)
      IF !Empty(oParent:freezeLeft)
        oParent:setLeftFrozen({})
      ENDIF
      oParent:delColumn(j)
    ENDIF
  ELSE
    lOk := .t.
    nInsColumn := 1
    FOR j := 1 TO oParent:colCount
      oColumn := oParent:getColumn(j)
      IF oColumn:columnPos == nColumn   // column is already in browse
        lOk := .f.
        EXIT
      ELSEIF oColumn:columnPos < nColumn
        nInsColumn := j
      ENDIF
    NEXT
    IF lOk
      IF nColumn > 1
        nInsColumn++
      ENDIF
      IF oParent:colCount == 0
        nInsColumn := 1
      ENDIF
      oParent:lockUpdate(.t.)
      IF !Empty(oParent:freezeLeft)
        oParent:setLeftFrozen({})
      ENDIF
      oParent:insColumn(nInsColumn, self)
      ::heading:useVisualStyle := ::visualStyleArray[1]
      ::dataArea:useVisualStyle := ::visualStyleArray[2]
      ::footing:useVisualStyle := ::visualStyleArray[3]
      ::heading:drawMode := ::drawModeArray[1]
      ::dataArea:drawMode := ::drawModeArray[2]
      ::footing:drawMode := ::drawModeArray[3]
    ENDIF
  ENDIF
  IF lOk
    IF !Empty(oParent:freezeLeft)
      oParent:setLeftFrozen(oParent:freezeLeft)
    ENDIF
    oParent:lockUpdate(.f.)
    oParent:forceStable()
    oParent:refreshAll()
    oParent:invalidateRect()
  ENDIF
ENDIF

IF Valtype(::whenBlock) = 'B'
  lEnable := Eval(::whenBlock,self)
  IF lEnable
    ::enable()
  ELSE
    ::disable()
  ENDIF
ENDIF

IF Valtype(lProtect) == 'L' .AND. Valtype(::protectBlock) == 'B'
  lProtect := Eval(::protectBlock,self)
ENDIF

RETURN self

* -----------------------

/*
METHOD DC_XbpColumn:WhenHideEval( lEnable, lShow, lProtect )

LOCAL oParent, nColumn, i, j, oColumn, nInsColumn, lOk, lUpdated := .f.

IF Valtype(::hideBlock) = 'B'
  oParent := ::parent
  nColumn := ::columnPos
  IF Eval(::hideBlock)
    lOk := .f.
    FOR j := 1 TO oParent:colCount
      oColumn := oParent:getColumn(j)
      IF oColumn:columnPos == nColumn // column has not been deleted
        lOk := .t.
        EXIT
      ENDIF
    NEXT
    IF lOk
      IF !Empty(oParent:freezeLeft)
        oParent:setLeftFrozen({})
      ENDIF
      oParent:delColumn(j)
    ENDIF
  ELSE
    lOk := .t.
    nInsColumn := 1
    FOR j := 1 TO oParent:colCount
      oColumn := oParent:getColumn(j)
      IF oColumn:columnPos == nColumn   // column is already in browse
        lOk := .f.
        EXIT
      ELSEIF oColumn:columnPos < nColumn
        nInsColumn := j
      ENDIF
    NEXT
    IF lOk
      IF nColumn > 1
        nInsColumn++
      ENDIF
      IF oParent:colCount == 0
        nInsColumn := 1
      ENDIF
      IF !Empty(oParent:freezeLeft)
        oParent:setLeftFrozen({})
      ENDIF
      oParent:insColumn(nInsColumn, self)
    ENDIF
  ENDIF
  IF lOk
    IF !Empty(oParent:freezeLeft)
      oParent:setLeftFrozen(oParent:freezeLeft)
    ENDIF
  ENDIF
ENDIF

IF Valtype(::whenBlock) = 'B'
  lEnable := Eval(::whenBlock,self)
  IF lEnable
    ::enable()
  ELSE
    ::disable()
  ENDIF
ENDIF

IF Valtype(lProtect) == 'L' .AND. Valtype(::protectBlock) == 'B'
  lProtect := Eval(::protectBlock,self)
ENDIF

RETURN self
*/

*----------------------

METHOD DC_XbpColumn:setHeaderCaption( bcHeader )

LOCAL i, aHeader

IF Valtype(bcHeader) = 'B'
  ::headerBlock := bcHeader
ELSEIF Valtype(bcHeader) $ 'CNO'
  ::headerCaption := bcHeader
ELSEIF !Empty(bcHeader)
  RETURN self
ENDIF

IF Valtype(::headerBlock) = 'B'
  bcHeader := DC_XtoC(Eval(::headerBlock))
ELSE
  bcHeader := ::headerCaption
ENDIF

::headerCaption := bcHeader
IF Valtype(::headerCaption) $ 'NO'
  ::headerCaption := DC_GetBitMap(::headerCaption,,,,,.t.)
  _LoadBitMap(::heading,::headerCaption)
ELSE
  aHeader := DC_TokenArray( bcHeader, ::parent:headerDelim )
  FOR i := 1 TO Len(aHeader)
    ::heading:setCell(i,aHeader[i])
  NEXT
  ::heading:invalidateRect()
ENDIF

RETURN self

* ---------------------

METHOD DC_XbpColumn:setHeaderColor( bColor )

LOCAL aColors

IF Empty(bColor) .AND. Valtype(::headerColorBlock) # 'B'
  RETURN self
ELSEIF Valtype(bColor) = 'B'
  ::headerColorBlock := bColor
ENDIF

aColors := Eval(::headerColorBlock)

IF Valtype(aColors) = 'A'
  ASize(aColors,2)
  IF Valtype(aColors[1]) == 'A'
    aColors[1] := GraMakeRGBColor(aColors[1])
  ENDIF
  IF Valtype(aColors[2]) == 'A'
    aColors[2] := GraMakeRGBColor(aColors[2])
  ENDIF
  ::heading:setColorFG(aColors[1])
  ::heading:setColorBG(aColors[2])
ENDIF

RETURN self

* ---------------------

METHOD DC_XbpColumn:setFooterColor( bColor )

LOCAL aColors

IF Empty(bColor) .AND. Valtype(::footerColorBlock) # 'B'
  RETURN self
ELSEIF Valtype(bColor) = 'B'
  ::footerColorBlock := bColor
ENDIF

aColors := Eval(::footerColorBlock)
IF Valtype(aColors) = 'A'
  ASize(aColors,2)
  IF Valtype(aColors[1]) == 'A'
    aColors[1] := GraMakeRGBColor(aColors[1])
  ENDIF
  IF Valtype(aColors[2]) == 'A'
    aColors[2] := GraMakeRGBColor(aColors[2])
  ENDIF
  ::footing:setColorFG(aColors[1])
  ::footing:setColorBG(aColors[2])
ENDIF

RETURN self

* ------------------

METHOD DC_XbpColumn:setFooterCaption( bcFooter )

LOCAL i, aFooter

IF Valtype(bcFooter) = 'B'
  ::footerBlock := bcFooter
ELSEIF Valtype(bcFooter) $ 'CNO'
  ::footerCaption := bcFooter
ELSEIF !Empty(bcFooter)
  RETURN self
ENDIF

IF Valtype(::footerBlock) = 'B'
  bcFooter := DC_XtoC(Eval(::footerBlock))
ELSE
  bcFooter := ::footerCaption
ENDIF

::footerCaption := bcFooter
IF Valtype(::footerCaption) $ 'NO'
  ::footerCaption := DC_GetBitMap(::footerCaption,,,,,.t.)
  _LoadBitMap(::footing,::footerCaption)
ELSE
  aFooter := DC_TokenArray( bcFooter, ::parent:footerDelim )
  FOR i := 1 TO Len(aFooter)
     ::footing:setCell(i,aFooter[i])
  NEXT
  ::footing:invalidateRect()
ENDIF

RETURN self

* ---------------------

METHOD DC_XbpColumn:MoveColumn( nMove )

LOCAL nPos, oBrowse

oBrowse := ::parent
oBrowse:lockUpdate(.t.)
nPos := oBrowse:colPos
oBrowse:delColumn(nPos)
oBrowse:insColumn(nPos+nMove,self)
IF nMove >= 1
  oBrowse:left()
ELSE
  oBrowse:right()
ENDIF
oBrowse:lockUpdate(.f.)
oBrowse:forceStable()
RETURN nil

* ----------------------

METHOD DC_XbpColumn:Sort( a, nColPos, nMode )

LOCAL i, j, oColumn, oXbp, aChildList, lStatus

DEFAULT nMode := 0

IF Valtype(::sortBlock) == 'B'
  IF !Empty(a)
    IF ::parent:dataSourceType # BROWSE_SQLEXPRESS .AND. ;
       (nColPos # ::parent:sortColumn .OR. !::parent:descendToggle)
      IF ::parent:descendToggle
        ::parent:descending := .f.
      ENDIF
      lStatus := Eval(::sortBlock,a,nColPos,::parent)
      IF Valtype(lStatus) = 'L' .AND. !lStatus
        RETURN self
      ENDIF
    ELSE
      IF nMode = 1
        ::parent:descending := .F.
      ELSEIF nMode = 2
        ::parent:descending := .T.
      ELSE
        ::parent:descending := !::parent:descending
        IF ::parent:descending
          ::parent:goBottom()
          ::parent:forceStable()
        ENDIF
      ENDIF
      if ::parent:dataSourceType = BROWSE_DATAOBJECT              // 04-01-2004 J.Duijf
         ::parent:datasource:DbSetDescend(::parent:descending)    // 04-01-2004 J.Duijf
         ::parent:datasource:DbGoTop()                            // 04-01-2004 J.Duijf
      elseif ::parent:dataSourceType = BROWSE_SQLEXPRESS
         Eval(::sortBlock,a,nColPos,::parent:descending)
      endif
    ENDIF
    IF Valtype(::parent:dataSource) $ 'AO'
      ::parent:goTop()
    ENDIF
    ::parent:refreshAll()
  ENDIF
ELSE
  RETURN self
ENDIF

FOR i := 1 TO ::parent:colCount
  oColumn := ::parent:GetColumn(i)
  IF i == nColPos
    IF nColPos # ::parent:sortColumn
      IF Valtype(oColumn:parent:sortColorSelectedFG) = 'N'
        oColumn:heading:setColorFG(oColumn:parent:sortColorSelectedFG)
      ENDIF
      IF Valtype(oColumn:parent:sortColorSelectedBG) = 'N'
        oColumn:heading:setColorBG(oColumn:parent:sortColorSelectedBG)
      ENDIF
    ENDIF
    IF oColumn:heading:status() > 0
      oXbp := nil
      aChildList := oColumn:heading:childList()
      FOR j := 1 TO Len(aChildList)
        IF aChildList[j]:isDerivedFrom('XbpStatic') .AND. ;
           Valtype(aChildList[j]:cargo) = 'C' .AND. ;
           aChildList[j]:cargo == 'SortArrow'
          oXbp := aChildList[j]
          EXIT
        ENDIF
      NEXT
      IF Empty(oXbp) .AND. ::parent:descendToggle
        oXbp := XbpStatic():new(oColumn:heading,, ;
                           {oColumn:heading:currentSize()[1]-15,6}, ;
                           {12,12},,.t.)
        oXbp:type := XBPSTATIC_TYPE_TEXT
        oXbp:cargo := 'SortArrow'
        oXbp:create()
        oXbp:setFontCompoundName('12.Marlett')
      ENDIF
      IF Valtype(oXbp) = 'O' .AND. oXbp:isDerivedFrom('XbpStatic') .AND. oXbp:type == XBPSTATIC_TYPE_TEXT
        IF ::parent:descending
          oXbp:setCaption('u')
        ELSE
          oXbp:setCaption('t')
        ENDIF
      ENDIF
    ENDIF
  ELSEIF nColPos # ::parent:sortColumn

    IF !Empty(oColumn:sortBlock)
      IF Valtype(oColumn:parent:sortColorUnSelectedFG) = 'N'
        oColumn:heading:setColorFG(oColumn:parent:sortColorUnSelectedFG)
      ENDIF
      IF Valtype(oColumn:parent:sortColorUnSelectedBG) = 'N'
        oColumn:heading:setColorBG(oColumn:parent:sortColorUnSelectedBG)
      ENDIF
    ELSE
      IF Valtype(oColumn:parent:sortColorNonselectableFG) = 'N'
        oColumn:heading:setColorFG(oColumn:parent:sortColorNonselectableFG)
      ENDIF
      IF Valtype(oColumn:parent:sortColorUnSelectedBG) = 'N'
        oColumn:heading:setColorBG(oColumn:parent:sortColorNonselectableBG)
      ENDIF
    ENDIF
    IF oColumn:heading:status() > 0 .AND. !Empty(oColumn:heading:childList())
      aChildList := oColumn:heading:childList()
      FOR j := 1 TO Len(aChildList)
        oXbp := aChildList[j]
        IF oXbp:isDerivedFrom('XbpStatic') .AND. Valtype(oXbp:cargo) = 'C' .AND. ;
           oXbp:cargo == 'SortArrow'
          // destroy the arrow
          oXbp:destroy()
        ENDIF
      NEXT
    ENDIF
  ENDIF
NEXT

::parent:sortColumn := nColPos

RETURN self
//
// J. Duijf 7-03-2010
//
// The methods below take care of the property ::Evenrow  .T. or .F.
// This can be used to inplement a zebra striping of the browse,
// even if browsing records in a database that contains deleted records.
//
METHOD DC_XbpColumn:DrawRow(nP1,xP2)
::LastRow      := nP1 + ::nOffset
::parent:evenrow := (Abs(::LastRow) % 2) = 0
RETURN ::XbpColumn:DrawRow(nP1,xP2)

METHOD DC_XbpColumn:ScrollUp(nP1,lP)
::nOffset      -= nP1
RETURN ::XbpColumn:ScrollUp(nP1,lP)

METHOD DC_XbpColumn:ScrollDown(nP1,lP)
::nOffset      += nP1
RETURN ::XbpColumn:ScrollDown(nP1,lP)

* ---------------------

STATIC FUNCTION _LoadBitmap( oParent, oCaption )

LOCAL oXbp

oXbp := XbpStatic():new(oParent,,{2,2},{oCaption:xSize,oCaption:ySize},,.t.)
oXbp:type := XBPSTATIC_TYPE_BITMAP
oXbp:caption := oCaption
oXbp:create()
DC_CenterObject(oXbp)

RETURN nil

* ----------------------

STATIC FUNCTION _HeaderDown( oBrowse, oColumn, nColPos )

DEFAULT nColPos := oBrowse:colCount

RETURN {|a,n,o|oColumn:Sort(a,nColPos)}

* --------------------

FUNCTION DC_SkipArray( nWantSkip, oBrowse )

LOCAL nDidSkip, nPointer, nLastRec := Len(oBrowse:dataSource), ;
      nSavePointer

IF nLastRec = 0
  RETURN 0
ENDIF

nDidskip := nWantSkip

IF Empty(oBrowse:filter)

  IF oBrowse:descending
    IF oBrowse:arrayElement - nWantSkip < 1 // bof
      nDidSkip := -1 * (1 - oBrowse:arrayElement)
    ELSEIF oBrowse:arrayElement - nWantSkip > nLastRec // eof
      nDidSkip := -1 * (nLastRec - oBrowse:arrayElement)
    ENDIF
    oBrowse:arrayElement -= nDidSkip
  ELSE
    IF oBrowse:arrayElement + nWantSkip < 1 // bof
      nDidSkip := 1 - oBrowse:arrayElement
    ELSEIF oBrowse:arrayElement + nWantSkip > nLastRec // eof
      nDidSkip := nLastRec - oBrowse:arrayElement
    ENDIF
    oBrowse:arrayElement += nDidSkip
  ENDIF

ELSE

  nDidSkip := 0
  nPointer := oBrowse:arrayElement

/*
  IF nPointer == 0
    oBrowse:arrayElement := 1
    oBrowse:cargo[4] := oBrowse:arrayElement
    RETURN 0
  ENDIF
*/

  nSavePointer := nPointer
  IF oBrowse:descending
    nWantSkip *= -1
  ENDIF
  IF nWantSkip == 0
    nDidSkip := 0
  ELSEIF nWantSkip > 0
    DO WHILE nDidSkip < nWantSkip
      nPointer++
      IF nPointer > Len(oBrowse:dataSource)
        nPointer := Len(oBrowse:dataSource)
        DO WHILE nPointer > 0 .AND. !Eval(oBrowse:filter,oBrowse:dataSource[nPointer])
          nPointer --
        ENDDO
        EXIT
      ELSEIF Eval(oBrowse:filter,oBrowse:dataSource[nPointer])
        nDidSkip++
      ENDIF
    ENDDO
  ELSEIF nWantSkip < 0
    DO WHILE nDidSkip > nWantSkip
      nPointer--
      IF nPointer < 1
        nPointer := 1
        DO WHILE nPointer < Len(oBrowse:dataSource) .AND. !Eval(oBrowse:filter,oBrowse:dataSource[nPointer])
          nPointer ++
        ENDDO
        EXIT
      ELSEIF Eval(oBrowse:filter,oBrowse:dataSource[nPointer])
        nDidSkip--
      ENDIF
    ENDDO
  ENDIF
  IF nDidSkip == 0
    nPointer := nSavePointer
  ENDIF
  oBrowse:arrayElement := nPointer

ENDIF

IF oBrowse:arrayElement > Len(oBrowse:dataSource)
  oBrowse:arrayElement := Len(oBrowse:dataSource)
ENDIF
oBrowse:cargo[4] := oBrowse:arrayElement

RETURN nDidSkip

* ------------------

STATIC FUNCTION ArrayPos( oBrowse )

RETURN Int(oBrowse:arrayElement/Len(oBrowse:dataSource))*100

* ------------------

STATIC FUNCTION ArrayGoPos( oBrowse, n )

RETURN Len(oBrowse:dataSource)*(n/100)


* ------------------

STATIC FUNCTION ArrayGoBottom( oBrowse )

LOCAL nElement := 0

IF Len(oBrowse:dataSource) = 0
  RETURN 0
ENDIF

IF Empty(oBrowse:filter)
  IF oBrowse:descending
    nElement := 1
  ELSE
    nElement := Len(oBrowse:dataSource)
  ENDIF
ELSE
  IF oBrowse:descending
    nElement := 1
    DO WHILE nElement <= Len(oBrowse:dataSource)
      IF Eval(oBrowse:filter,oBrowse:dataSource[nElement])
        EXIT
      ENDIF
      nElement++
    ENDDO
  ELSE
    nElement := Len(oBrowse:dataSource)
    DO WHILE nElement > 0
      IF Eval(oBrowse:filter,oBrowse:dataSource[nElement])
        EXIT
      ENDIF
      nElement--
    ENDDO
  ENDIF
ENDIF

RETURN nElement

* ------------------

STATIC FUNCTION ArrayGoTop( oBrowse )

LOCAL nElement := 0

IF Len(oBrowse:dataSource) = 0
  RETURN 0
ENDIF

IF Empty(oBrowse:filter)
  IF oBrowse:descending
    nElement := Len(oBrowse:dataSource)
  ELSE
    nElement := 1
  ENDIF
ELSE
  IF !oBrowse:descending
    nElement := 1
    DO WHILE nElement <= Len(oBrowse:dataSource)
      IF Eval(oBrowse:filter,oBrowse:dataSource[nElement])
        EXIT
      ENDIF
      nElement++
    ENDDO
    IF nElement > Len(oBrowse:dataSource)
      nElement := 0
    ENDIF
  ELSE
    nElement := Len(oBrowse:dataSource)
    DO WHILE nElement > 0
      IF Eval(oBrowse:filter,oBrowse:dataSource[nElement])
        EXIT
      ENDIF
      nElement--
    ENDDO
  ENDIF
ENDIF

RETURN nElement

* ------------------

STATIC FUNCTION _Pres( aPres )

IF Valtype( aPres ) = 'B'
  RETURN Eval(aPres)
ENDIF

RETURN aPres

* ----------------------

STATIC FUNCTION _ColPicture( bBlock, bcPicture, oColumn, oRecord )

LOCAL bDataLink

IF Valtype(bcPicture) = 'B'

  IF Valtype(oRecord) == 'O'
    bDataLink := {|x,y,z|x := Eval(bBlock,oRecord,oColumn), ;
                       z := Eval(bcPicture), ;
                       y:=IIF( x==NIL,Space(Len(z)),Transform(x,z)),y }
  ELSE
    bDataLink := {|x,y,z|x := Eval(bBlock,x,oColumn), ;
                         z := Eval(bcPicture), ;
                         y:=IIF( x==NIL,Space(Len(z)),Transform(x,z)),y }
  ENDIF

ELSE

  IF Valtype(oRecord) == 'O'
    bDataLink := {|x,y,z|x := Eval(bBlock,oRecord,oColumn), ;
                         z := bcPicture, ;
                         y:=IIF( x==NIL,Space(Len(z)),Transform(x,z)),y }
  ELSE
    bDataLink := {|x,y,z|x := Eval(bBlock,x,oColumn), ;
                       z := bcPicture, ;
                       y:=IIF( x==NIL,Space(Len(z)),Transform(x,z)),y }
  ENDIF

ENDIF

RETURN bDataLink

* ----------------------

STATIC FUNCTION _ColorBlock( aColor )

IF Valtype(aColor) = 'A' .AND. Len(aColor) = 2
  IF Valtype(aColor[1]) = 'B'
     RETURN aColor[1]
  ELSE
     RETURN {||aColor}
  ENDIF
ENDIF

RETURN nil

* ---------------------

STATIC PROCEDURE _Optimize( a, b, o )

LOCAL aArray, oColumn := o:setParent(), i, nWidth := 0, ;
      oPS, cCaption, oBrowse, aGraArray, nPointer, bDataLink, ;
      nColumn, cAlias, nSaveRec

Grok(a,b)
oBrowse  := oColumn:parent
nColumn := oBrowse:colPos

nPointer := oBrowse:arrayElement  // save pointer
bDataLink := oColumn:dataLink

oPS := o:lockPS()

IF Empty(oColumn:columnSize)
  oColumn:columnSize := oColumn:currentSize()[1]
  IF Valtype(oColumn:dataSource) = 'A'
    aArray   := oColumn:dataSource
    FOR i := 1 TO Len(aArray)
      oBrowse:arrayElement := i
      cCaption := DC_XtoC(Eval(bDataLink))
      aGraArray := GraQueryTextBox( oPS, cCaption )
      nWidth := Max( nWidth, ( aGraArray[3,1] - aGraArray[1,1]) + 2 )
    NEXT
  ELSEIF Valtype(oColumn:dataSource) = 'C'
    cAlias := oColumn:dataSource
    nSaveRec := (cAlias)->(RecNo())
    (cAlias)->(DC_DbGoTop())
    DO WHILE !(cAlias)->(DC_Eof())
      cCaption := (cAlias)->(DC_XtoC(Eval(bDataLink)))
      aGraArray := GraQueryTextBox( oPS, cCaption )
      nWidth := Max( nWidth, ( aGraArray[3,1] - aGraArray[1,1]) + 2 )
      (cAlias)->(DC_DbSkip())
    ENDDO
    (cAlias)->(dbGoTo(nSaveRec))
  ELSE
    o:unlockPS()
    RETURN
  ENDIF
ELSE
  nWidth := oColumn:columnSize
  oColumn:columnSize := 0
ENDIF

oBrowse:arrayElement := nPointer // restore pointer

IF Valtype(Eval(bDataLink)) = 'D'
  nWidth += 10
ELSEIF Valtype(Eval(bDataLink)) = 'N'
  nWidth += 30
ELSE
  nWidth += 10
ENDIF

o:unlockPS()
oColumn:setSize( {nWidth, oColumn:currentSize()[2] })
oBrowse:setSize(oBrowse:currentSize())
oBrowse:firstCol()
FOR i := 1 TO nColumn-1
  oBrowse:right()
NEXT
oBrowse:forceStable()

RETURN

* ---------------

FUNCTION DC_GetColArray( nElement, oBrowse, xVar )

LOCAL aArray, nPointer, xValue

aArray   := oBrowse:dataSource
nPointer := oBrowse:arrayElement

IF nPointer > Len(aArray) .OR. nPointer = 0
  RETURN ''
ENDIF

IF nElement == NIL
  RETURN aArray[nPointer]
ELSEIF nElement <= 0
  RETURN nPointer
ENDIF

xValue := aArray[nPointer,nElement]

IF Valtype(xVar) == Valtype(xValue)
  aArray[nPointer,nElement] := xVar
ENDIF

RETURN aArray[nPointer,nElement]

* -------------------

FUNCTION DC_BrowseRow( oBrowse, nElement )

LOCAL aArray, nOldElement

IF Valtype(oBrowse) # 'O'
   RETURN 0
ENDIF

aArray   := oBrowse:cargo[5]
nOldElement := oBrowse:cargo[4]

IF Valtype(nElement) = 'N' .AND. nElement > 0 .AND. nElement <= Len(aArray)
  oBrowse:cargo[4] := nElement
  oBrowse:arrayElement := nElement
ENDIF

RETURN nOldElement

* ----------------------

FUNCTION DC_GetBrowArray( oBrowse, aArray )

IF Valtype( aArray ) = 'A'
  IF oBrowse:isDerivedFrom('XbpBrowse')
    oBrowse:cargo[5] := aArray
  ENDIF
  oBrowse:dataSource := aArray
ENDIF

RETURN oBrowse:dataSource

* ----------------------

FUNCTION DC_BrowseSort( aSort )

STATIC saSort := ;
   { GRA_CLR_WHITE, ;    // Sort Selected Color (Foreground)
     GRA_CLR_RED, ;      // Sort Selected Color (Background)
     nil, ;              // Sort Nonselectable Color (Foreground)
     nil, ;              // Sort Nonselectable Color (Background)
     nil, ;              // Sort UP Bitmap (no longer used)
     nil, ;              // Sort DOWN Bitmap (no longer used)
     GRA_CLR_WHITE, ;    // Sort Unselected Color (Foreground)
     GRA_CLR_BLUE }      // Sort Unselected Color (Background)

LOCAL aOldSort := saSort

IF Valtype(aSort) = 'A'
  saSort := aSort
ENDIF

DEFAULT aOldSort := {}
ASize(aOldSort,8)

RETURN aOldSort

* ----------------------

FUNCTION DC_BrowseColor( aColor )

STATIC saColor := nil

LOCAL aOldColor := saColor

IF Valtype(aColor) = 'A'
  saColor := aColor
  ASize(saColor,4)
  IF Valtype(saColor[1]) == 'A'
    saColor[1] := GraMakeRGBColor(saColor[1])
  ENDIF
  IF Valtype(saColor[2]) == 'A'
    saColor[2] := GraMakeRGBColor(saColor[2])
  ENDIF
  IF Valtype(saColor[3]) == 'A'
    saColor[3] := GraMakeRGBColor(saColor[3])
  ENDIF
  IF Valtype(saColor[4]) == 'A'
    saColor[4] := GraMakeRGBColor(saColor[4])
  ENDIF
ELSEIF Pcount() == 1 .AND. aColor == NIL
  saColor := NIL
ENDIF

RETURN aOldColor
FUNCTION DC_MergePresentation(aTarget,aSource )
//
// J. Duijf    10-03-2011
//
// Merge presentation parameters.
// Both parameters are 2 dimensional arrays.
// Add item to target in not exist.
// Replace value is already exist.
//
// This function is used while creating Dc_XbpBrowse and Dc_XbpColumn
//
LOCAL nPos           := 0
LOCAL ni             := 0
if ValType(aTarget) = "A" .and. ;                                             // Make sure its a array
   ValType(aSource) = "A"
   For ni := 1 To Len(aSource)
      nPos           := Ascan(aTarget,{|a|a[1] = aSource[ni,1]})              // Locate property in target
      if nPos = 0                                                             // Not there,
         Aadd(aTarget,Aclone(aSource[ni]))                                    // Add to target
      else                                                                    // Ok, found
         aTarget[nPos,2] := aSource[ni,2]                                     // Replace propperty
      endif
   Next ni
endif
Return aTarget                                                                // Updated array presentation parameters

FUNCTION DC_GetPresentation(aPres,nId,uDefault)
//
// J. Duijf    10-03-2011
//
// Locate ID in presentation parameters.
// Return value if found, or default value if not found
//
// This function is used while creating Dc_XbpBrowse and Dc_XbpColumn
//
LOCAL nPos           := 0
LOCAL uRet           := uDefault
if ValType(aPres) = "A"                                                       // Make sure its a array
   nPos              := Ascan(aPres,{|a|a[1] == nId})
   if nPos > 0                                                                // ID found
      uRet           := aPres[nPos,2]                                         // Return value
   endif
endif
Return uRet

* ------------

CLASS DC_XbpCellGroupMultiLine FROM XbpCellGroup

PROTECTED:
   VAR aRows

EXPORTED:
   VAR column
   VAR browse
   VAR dataLinkMultiLine

   METHOD setCell
   METHOD drawCell
   METHOD hiliteCell
   METHOD paint
   METHOD setSize

   INLINE METHOD destroy
      ::XbpCellGroup:destroy()
//      ::aRows:= ::column:= ::browse := nil
   RETURN self

ENDCLASS

* -------------

METHOD DC_XbpCellGroupMultiLine:setCell( nRowPos, xValue, nType, lRepaint )

LOCAL aData

IF xValue <> nil
   aData:= {}
   AEVAL( ::dataLinkMultiLine, {|b| AADD( aData, EVAL( b ) ) } )
ENDIF

RETURN ::XbpCellGroup:setCell( nRowPos, aData, nType, lRepaint )

* --------------

METHOD DC_XbpCellGroupMultiLine:hiliteCell( nRowPos, lHilite, lFrame, lRepaint )

LOCAL oPS, aAttr, aAttrAntes, nColorHiliteFG, nFound, aPP
LOCAL aData := ::getCell( nRowPos )
LOCAL lExit := .T.
//
// J. Duijf 19-02-2012
// To prevent garbled data on the first line of a multiline browse:
// Clear the cell data just before highliting
//
if ValType(aData) = "A"                                                       // Only if multiline cell
   ::XbpCellGroup:setCell( nRowPos, "",,.F.)                                  // Clear cell (no paint)
   lExit := ::XbpCellGroup:hiliteCell( nRowPos, lHilite, lFrame, lRepaint )   // Hilite
   ::XbpCellGroup:setCell( nRowPos, aData,,.F.)                               // Restore cell data (no paint)
else                                                                          // Regular data
   lExit := ::XbpCellGroup:hiliteCell( nRowPos, lHilite, lFrame, lRepaint )
endif
IF aData <> nil

   oPS := ::lockPS()

   IF lHilite
     nColorHiliteFG := XBPSYSCLR_HILITEBACKGROUND
     aPP := ::browse:setPresParam()
     nFound := AScan(aPP,{|a|a[1]==XBP_PP_COL_DA_HILITE_FGCLR})
     IF nFound = 0
      nFound := AScan(aPP,{|a|a[1]==XBP_PP_HILITE_FGCLR})
     ENDIF
     IF nFound > 0
       nColorHiliteFG := aPP[nFound,2]
     ENDIF
     aAttr:= ARRAY( GRA_AS_COUNT )
     aAttr[ GRA_AS_COLOR ] := nColorHiliteFG
     aAttrAntes:= oPS:SetAttrString( aAttr )
     _DrawCell( self, oPS, nRowPos )
     oPS:SetAttrString( aAttrAntes )
   ELSE
     _DrawCell( self, oPS, nRowPos )
   ENDIF

   ::unLockPs( oPS )

ENDIF

RETURN lExit

* -------------

METHOD DC_XbpCellGroupMultiLine:drawCell( xRowPos, lRepaint )

LOCAL oPS, nCont, nFin, lExit:= ::XbpCellGroup:drawCell( xRowPos, lRepaint )

oPS:= ::lockPS()

IF ValType( xRowPos ) == 'A'
    nFin:= LEN( xRowPos )
    FOR nCont:= 1 TO nFin
      IF xRowPos[ nCont ] <> nil
        _DrawCell( self, oPS, xRowPos[ nCont ] )
      ENDIF
    NEXT
ELSEIF ValType( xRowPos ) == 'N'
    _DrawCell( self, oPS, xRowPos )
ENDIF

::unLockPs( oPS )

RETURN lExit

* --------------

METHOD DC_XbpCellGroupMultiLine:paint( aRect )

Grok(aRect)
IF NextAppEvent() <> xbeBRW_Navigate

   ::drawCell( ::aRows, .T. )
   IF ::browse:cursorMode == XBPBRW_CURSOR_ROW ;
         .OR. ( ::browse:cursorMode == XBPBRW_CURSOR_CELL ;
                .AND. ::browse:getColumn( ::browse:colPos ) == ::column )
      ::hiliteCell( ::browse:rowPos, .T., .F., .T. )
   ENDIF

ENDIF

RETURN self

* -----------

METHOD DC_XbpCellGroupMultiLine:setSize( aSize, lPaint )

LOCAL lExit := ::XbpCellGroup:setSize( aSize, lPaint )

::aRows:= AEVAL( Array( ::rowCount() ), {|x,i| x:= i },,, .T. )

RETURN lExit

*-------------

STATIC FUNCTION _DrawCell( oCellGroup, oPS, nRowPos )

LOCAL self := oCellGroup, aPos, nY, aArray, nSpace, ;
      nCellHeight, nFontHeight, nAlign, nFound, nOffset, aPP, ;
      aGetData[0], aData[0], i, j

aGetData := ::getCell(nRowPos)

IF aGetData <> nil

  FOR i := 1 TO Len(aGetData)
    IF Valtype(aGetData[i]) = 'B'
      AAdd(aData,Eval(aGetData[i]))
    ELSEIF Valtype(aGetData[i]) == 'A'
      FOR j := 1 TO Len(aGetData[i])
        AAdd(aData,aGetData[i,j])
      NEXT
    ELSE
      AAdd(aData,aGetData[i])
    ENDIF
  NEXT

  aPos:= ::cellRect( nRowPos, .T. )

  nCellHeight := aPos[4]-aPos[2]
  aArray := GraQueryTextBox( oPS, 'test' )
  nFontHeight := aArray[1,2] - aArray[2,2] // height

  aPP := ::column:presArray
  nFound := AScan(aPP,{|a|Valtype(a[1])=='N'.AND. ;
                   (a[1]==XBPCOL_DA_CELLALIGNMENT .OR. ;
                    a[1]==XBP_PP_COL_DA_CELLALIGNMENT) } )

  IF nFound == 0
    aPP := ::browse:setPresParam()
    nFound := AScan(aPP,{|a|Valtype(a[1])=='N'.AND. ;
                   a[1]==XBP_PP_COL_DA_CELLALIGNMENT} )
  ENDIF
  IF nFound > 0
    nAlign := aPP[nFound,2]
  ELSE
    nAlign := XBPALIGN_VCENTER
  ENDIF

  nSpace := nFontHeight
  nOffset := 3
  IF nSpace * Len(aData) >= nCellHeight + nOffset
    IF nAlign == XBPALIGN_VCENTER
      nOffset := 0
    ENDIF
    nSpace := nCellHeight/Len(aData)
  ENDIF

  IF nAlign = XBPALIGN_TOP
    nY := aPos[2] + nCellHeight - nFontHeight + nOffset
  ELSEIF DC_BitTest(nAlign,4) // align bottom
    nY := aPos[2] +  ( nSpace * Len(aData) ) - nSpace
  ELSEIF DC_BitTest(nAlign,3) // align vcenter
    nY := aPos[2] + ;
      ( nCellHeight + ( nSpace * Len(aData) ) ) / 2 - nSpace + nOffset
  ELSE // align top
    nY := aPos[2] + nCellHeight - nFontHeight + nOffset
  ENDIF

  AEVAL( aData, {|xDat| GraStringAt( oPS, ;
           { _AlignText(aPos,oPS,nAlign,xDat), nY }, Var2LChar( xDat ) ), nY -= nSpace } )
ENDIF

RETURN nil

* ------------

STATIC FUNCTION _AlignText(aPos,oPS,nAlign,xDat)

LOCAL nCellWidth, nTextWidth, aArray

IF DC_BitTest(nAlign,2) // align right
  nCellWidth := aPos[3] - aPos[1]
  aArray := GraQueryTextBox(oPS,xDat)
  nTextWidth := aArray[3,1] - aArray[1,1]
  RETURN nCellWidth - nTextWidth
ELSEIF DC_BitTest(nAlign,1) // align hcenter
  nCellWidth := aPos[3] - aPos[1]
  aArray := GraQueryTextBox(oPS,xDat)
  nTextWidth := aArray[3,1] - aArray[1,1]
  RETURN (nCellWidth - nTextWidth) / 2
ENDIF

RETURN aPos[1]

*-------------

STATIC FUNCTION ACEDbSkipper( nWantSkip, oServer )

   LOCAL nDidSkip := 0

   DO CASE
   CASE oServer:LastRec() == 0
   CASE nWantSkip == 0
      oServer:Skip(0)

   CASE nWantSkip > 0 .AND. oServer:Recno() <> oServer:LastRec() + 1

      // Abwrts skippen
      DO WHILE nDidSkip < nWantSkip
         oServer:Skip(1)
         IF oServer:Eof()
            IF .f. // lAppend
               // Append Modus: Ghost record
               nDidSkip ++
            ELSE
               oServer:Skip(-1)
            ENDIF
            EXIT
         ENDIF
         nDidSkip ++
      ENDDO

   CASE nWantSkip < 0

      // Aufwrts skippen
      DO WHILE nDidSkip > nWantSkip
         oServer:Skip(-1)
         IF oServer:Bof()
            EXIT
         ENDIF
         nDidSkip --
      ENDDO
   ENDCASE

RETURN  nDidSkip

* --------------

FUNCTION DC_AutoRestoreBrowse( aParams )

STATIC saParams := nil
LOCAL aOldParams := saParams

// [1] nHKey
// [2] Start Folder
// [3] Save Proc Name

IF Valtype(aParams) = 'A'
  saParams := aParams
  ASize(saParams,3)
  DEFAULT saParams[3] := .T.
ENDIF

RETURN aOldParams

* -------------

STATIC FUNCTION ApplyScaleFactor( aPres, oGetList )

LOCAL nScale, i

IF Valtype(aPres) = 'A'
  IF (nScale := oGetList:scaleFactor[4]) # 1
    aPres := AClone(aPres)
    FOR i := 1 TO Len(aPres)
      IF aPres[i,1] $ { XBP_PP_COL_DA_ROWHEIGHT, XBP_PP_COL_HA_HEIGHT, XBP_PP_COL_FA_HEIGHT }
         If aPres[i,2] <> -1
            aPres[i,2] *= nScale
         Endif
      ENDIF
    NEXT
  ENDIF
ENDIF

RETURN aPres

* --------------

FUNCTION DC_BrowseSQL( oCursor )

LOCAL GetList[0], oBrowse, i, GetOptions

IF Valtype(oCursor) # 'O'
  RETURN .F.
ENDIF

@ 0,0 DCBROWSE oBrowse DATA oCursor ;
  SIZE 100,25 ;
  ID 'SQL_BROWSE' ;
  PRESENTATION DC_BrowPres() ;
  CURSORMODE XBPBRW_CURSOR_ROW

FOR i := 1 TO oCursor:fCount()

  DCBROWSECOL DATA SQLFieldBlock( @oCursor, i ) ;
     WIDTH 10 ;
     HEADER oCursor:fieldName(i) ;
     PARENT oBrowse ;
     SORT SQLSortBlock(@oBrowse,i)

NEXT

DCGETOPTIONS ;
   RESIZE ;
   RESIZEDEFAULT DCGUI_RESIZE_RESIZEONLY

DCREAD GUI ;
   FIT ;
   OPTIONS GetOptions ;
   TITLE 'SQL Browse'

RETURN .T.

* ---------------

STATIC FUNCTION SQLFieldBlock( oCursor, nField )

Return {||oCursor:fieldGet(nField)}

* ---------------

STATIC FUNCTION SQLSortBlock( oBrowse, nField, nColumn )

RETURN {|a,b,descend|;
   IIF(valtype(nField)=='C',nField:=oBrowse:dataSource:fieldPos(nField),nil), ;
       oBrowse:dataSource:sort(IIF(descend,;
       {|x,y|IIF(Valtype(x[nField])='C', ;
             Upper(x[nField]) > Upper(y[nField]), x[nField] > y[nField])}, ;
       {|x,y|IIF(Valtype(x[nField])='C', ;
             Upper(x[nField]) < Upper(y[nField]), x[nField] < y[nField])})), ;
       nColumn := nField}


* --------------

FUNCTION DC_XbpBrowseDefault( oBrowDefault )

STATIC soBrowDefault := nil
LOCAL oOldBrowDefault

oOldBrowDefault := soBrowDefault

IF Valtype(oBrowDefault) == 'O'
  soBrowDefault := oBrowDefault
ENDIF

IF Empty(oOldBrowDefault)
  oOldBrowDefault := DC_XbpBrowseConfig():new()
ENDIF

RETURN oOldBrowDefault

* ----------------

CLASS DC_XbpBrowseConfig

EXPORTED:

VAR font
VAR presParam
VAR useVisualStyleBrowse
VAR useVisualStyleScrollBars

METHOD Init

ENDCLASS

* ---------------

METHOD DC_XbpBrowseConfig:init()

::font := nil
::presParam := DC_BrowPres()
::useVisualStyleBrowse := .f.
::useVisualStyleScrollBars := .t.

RETURN self

* ---------------

GETSETFUNCTION DC_GetWhenOnBrowseItemMarked DEFAULT .f.

* ---------------

FUNCTION DC_Browse2Dbf( oBrowse, cDbfName, aStru )

LOCAL oColumn, aColumns, i, nRecord, nPreviousRecord, xValue, lStatus := .t., ;
      cFiblockpe, cFieldName, cFieldType, cAlias

IF Valtype(oBrowse:dataSource) = 'C'
  nPreviousRecord := (oBrowse:dataSource)->(RecNo())
  oBrowse:stableblock := {||nRecord := (oBrowse:dataSource)->(RecNo())}
ELSEIF Valtype(oBrowse:dataSource) = 'A'
  nPreviousRecord := oBrowse:arrayElement
  oBrowse:stableblock := {||nRecord := oBrowse:arrayElement}
ELSE
  RETURN .f.
ENDIF

aColumns := Array(oBrowse:colCount)

IF Len(aStru) # Len(aColumns)
  RETURN .f.
ENDIF

dbCreate( cDbfName, aStru )
USE (cDbfName) NEW
cAlias := Alias()

oBrowse:goTop()

FOR i := 1 TO Len(aColumns)
  oColumn := oBrowse:getColumn(i)
  aColumns[i]   := oColumn
NEXT

DO WHILE .t.

  (cAlias)->(DC_AddRec())

  FOR i := 1 TO Len(aColumns)
    xValue := Eval(aColumns[i]:dataLink)
    cFieldName := aStru[i,1]
    cFieldType := aStru[i,2]
    IF !(Valtype(xValue) == cFieldType)
      IF cFieldType == 'N' .AND. Valtype(xValue) == 'C'
        xValue := Val(xValue)
      ELSEIF cFieldType == 'D' .AND. Valtype(xValue) == 'C'
        xValue := Ctod(xValue)
      ELSEIF cFieldType = 'C'
        xValue := DC_XtoC(xValue)
      ENDIF
    ENDIF
    REPLACE (cAlias)->&(cFieldName) WITH xValue
  NEXT
  (cAlias)->(dbRUnlock())

  oBrowse:down()
  oBrowse:forcestable()
  IF nRecord == nPreviousrecord
    EXIT
  ENDIF
  nPreviousrecord := nRecord

ENDDO

RETURN lStatus

* -----------

FUNCTION DC_ObjectVarBlock( cVar, nElement )

LOCAL bAccess

bAccess := {|u,o|_ObjectAccess(u,o,cVar,nElement)}

RETURN bAccess

* ---------

STATIC FUNCTION _ObjectAccess( u,oColumn,cVar,nElement )

LOCAL oBrowse, aData, oObject, bError, lStatus := .f.
LOCAL oError
IF Valtype(oColumn) # 'O'
  RETURN ''
ENDIF

oBrowse := oColumn:parent
aData := oBrowse:dataSource
IF Empty(aData)
   //
   // Return correct default value
   // Check the column type
   // J. Duijf 26-07-2014
   //
   // This prevents runtime error if browse is array, and contains images
   //
   Do Case
      Case oColumn:Type = XBPCOL_TYPE_TEXT
         RETURN ''
      Case oColumn:Type = XBPCOL_TYPE_MULTILINETEXT
         RETURN ''
      Otherwise
//         XBPCOL_TYPE_ICON
//         XBPCOL_TYPE_BITMAP
//         XBPCOL_TYPE_SYSICON
//         XBPCOL_TYPE_FILEICON
//         XBPCOL_TYPE_FILEMINIICON
         Return 0
   Endcase
ENDIF
//
// 12-12-2013 J. Duijf
// Prevent runtime error. Test array boundary
//
If oBrowse:arrayElement <= Len(aData) .and. oBrowse:arrayElement > 0
   IF Valtype(nElement) == 'N'
     If nElement <= Len(aData[oBrowse:arrayElement]) .and. nElement > 0
        oObject := aData[oBrowse:arrayElement,nElement]
     Endif
   ELSE
     oObject := aData[oBrowse:arrayElement]
   ENDIF
Endif
IF Valtype(oObject) # 'O'
  u := '<not an object>'
  lStatus := .t.
ELSE
  PRIVATE _oObject := oObject
  PRIVATE _uData   := u
//  bError := ErrorBlock({|e|Break(e)})
  BEGIN SEQUENCE
  IF '{' $ cVar
    cVar := &cVar
  ENDIF
  IF u == nil
    IF Valtype(cVar) = 'B'
      u := Eval(cVar,oObject)
    ELSE
      u := &('_oObject:' + cVar)
    ENDIF
  ELSE
    IF Valtype(cVar) = 'B'
      u := Eval(cVar,oObject,u)
    Else
      u := &('_oObject:' + cVar + " := _uData")
    Endif
  ENDIF
  lStatus := .t.
  RECOVER using oError
   oError := oError
  END SEQUENCE
//  ErrorBlock(bError)
  Release _oObject, _uData
ENDIF
IF !lStatus
  u := '<' + DC_XtoC(cVar) + ':undefined>'
ENDIF
RETURN u

* ---------------

CLASS DC_BrowseQuery

EXPORTED:

VAR alias
VAR queryArray
VAR browseObject
VAR browseObjectBlock
VAR parentObject
VAR queryGetList
VAR mainGetList
VAR buttonConfig
VAR startBlock
VAR doneBlock
VAR bgColor
VAR staticRow
VAR staticCol
VAR staticWidth
VAR staticHeight
VAR isPixel
VAR isSQL
VAR saySize
VAR isFloating
VAR isVertical
VAR dialogWindow
VAR maxGetWidth

METHOD init, buildGets, applyQuery, testQuery, clearFields, isVisible, destroy

ENDCLASS

* ----------

METHOD DC_BrowseQuery:init( GetList )

::isSQL := .f.
::isFloating := .f.
::isVertical := .f.
::alias := Alias()
::mainGetList := GetList
::saySize := 15
::staticRow := 0
::staticCol := 0
::maxGetWidth := 100

RETURN self

* ----------

METHOD DC_BrowseQuery:buildGets( oBrowse, oStaticBrowse )

LOCAL i, nRow := .5, nCol := 0, dPopUp, nSaveRec := (::alias)->(RecNo()), lData := .F., ;
      nColOffset, nButtonCol, GetList, oStaticQuery, oStaticText, nWidth, GetOptions, ;
      nGetSize

nButtonCol := 1

IF ::isFloating
  GetList := {}
  oStaticBrowse := nil
ELSE
  GetList := ::mainGetList
  ::isVertical := .f.
ENDIF
::queryGetList := GetList

::browseObjectBlock := DC_GetAnchorCb(@oBrowse)

(::alias)->(dbGoto(0))

IF !::isFloating

  @  ::staticRow,  ::staticCol DCSTATIC TYPE XBPSTATIC_TYPE_RECESSEDBOX OBJECT oStaticQuery ;
                SIZE ::staticWidth, ::staticHeight PARENT oStaticBrowse ;
                EVAL {|o|IIF(!::isFloating,o:hide(),nil)}

  @   2,  2 DCSTATIC TYPE XBPSTATIC_TYPE_TEXT OBJECT oStaticText ;
                SIZE DCGUI_PARENTWIDTH-4, DCGUI_PARENTHEIGHT-4 ;
                PARENT oStaticQuery COLOR GRA_CLR_BLACK, ::bgColor ;
                PIXEL

ENDIF

nRow := .5

@ nRow, nButtonCol DCPUSHBUTTON CAPTION 'Find' SIZE 8,1 ;
         ACTION {||::applyQuery(@oBrowse)} PARENT oStaticText GROUP 'QUERYBUTTONS' ;
         WHEN {|| ::testQuery() .OR. lData } ;
         TOOLTIP "Find Records Matching The Selected Search Criteria" ;
         CONFIG ::buttonConfig

IF !::isVertical
  nRow++
ELSE
  nButtonCol += 10
ENDIF

@ nRow, nButtonCol DCPUSHBUTTON CAPTION 'Clear' SIZE 8,1 ;
         ACTION {||::clearFields(@oBrowse)} PARENT oStaticText GROUP 'QUERYBUTTONS' ;
         WHEN {|| ::testQuery() .OR. lData } ;
         TOOLTIP "Reset Search Criteria" ;
         CONFIG ::buttonConfig

IF !::isVertical
  nRow++
ELSE
  nButtonCol += 10
ENDIF

@ nRow, nButtonCol DCPUSHBUTTON CAPTION 'Done' SIZE 8,1 ;
         ACTION {||::clearFields(@oBrowse), ;
                   Eval(::doneBlock)} PARENT oStaticText ;
         TOOLTIP "Close Search Window" ;
         CONFIG ::buttonConfig

IF !::isVertical
  nColOffset := nButtonCol + 10
  nRow := .5
ELSE
  nColOffset := 0
  nRow += 2
ENDIF

nCol := 0
nWidth := 0

DEFAULT ::staticHeight := (Len(::queryArray) / 3 + 1)

FOR i := 1 TO Len(::queryArray)
  ASize(::queryArray[i], QUERY_VAR)
  ::queryArray[i,QUERY_ROW] := nRow
  ::queryArray[i,QUERY_COL] := nCol
  ::queryArray[i,QUERY_GET] := (::alias)->(&(::queryArray[i,QUERY_FIELD]))

  nGetSize := Len(DC_XtoC(::queryArray[i,QUERY_GET]))
  nGetSize := Min(nGetSize,::maxGetWidth)

  nWidth := Max( nWidth, nGetSize )

  nRow++
  IF !::isVertical .AND. nRow > ::staticHeight-1
    nRow := .5
    nCol += ::saySize + nWidth + 2
    nWidth := 0
  ENDIF
NEXT

FOR i := 1 TO Len(::queryArray)

  dPopup := nil
  nGetSize := nil
  IF Valtype(::queryArray[i,QUERY_GET]) == 'D'
    dPopup := {|d|DC_PopDate(d,,,,,,2)}
  ELSEIF Valtype(::queryArray[i,QUERY_GET]) == 'C'
    nGetSize := Len(::queryArray[i,QUERY_GET])
    nGetSize := Min(nGetSize,::maxGetWidth)
  ENDIF

  nRow := IIF( ValType(::queryArray[i,QUERY_ROW]) == 'N', ::queryArray[i,QUERY_ROW], nRow++ )
  nCol := IIF( ValType(::queryArray[i,QUERY_COL]) == 'N', ::queryArray[i,QUERY_COL], 0 ) + nColOffset

  @ nRow++, nCol DCSAY ::queryArray[i,QUERY_SAY] GET ::queryArray[i,QUERY_GET] ;
                 KEYBLOCK {|a,b,o| IIF( Len(AllTrim(o:Get:buffer)) <= 2, ;
                                       ( lData := !Empty(AllTrim(o:Get:buffer)), DC_GetRefresh(::queryGetList,,,,{'QUERYBUTTONS'})), nil )} ;
                 SAYSIZE ::saySize SAYRIGHTBOTTOM TABSTOP PARENT oStaticText POPUP dPopUp ;
                 GETCOLOR GRA_CLR_BLACK, GRA_CLR_WHITE ;
                 GETID 'QUERY_' + Alltrim(Str(i)) ;
                 GETEVAL _GetEvalBlock(::queryArray[i],ATail(::queryGetList)) ;
                 PICTURE ::queryArray[i,QUERY_PICTURE] ;
                 GETSIZE nGetSize

NEXT

DCEVAL {||::browseObject := oBrowse, ;
          ::parentObject := oStaticQuery}

IF ::isFloating

  ::startBlock := {||::dialogWindow:show(), ;
                    ::dialogWindow:toFront(), ;
                    SetAppFocus(DC_GetObject(::queryGetList,'QUERY_1')), ;
                    DC_GetRefresh(::mainGetList)}

  ::doneBlock := {||::dialogWindow:hide(), ;
                    DC_GetRefresh(::mainGetList)}

ELSE

  ::startBlock := {|| IIF( oStaticQuery:isVisible(), nil, ;
                         (oBrowse:setSize( {oBrowse:currentSize()[1],oBrowse:currentSize()[2]-oStaticQuery:currentSize()[2]}), ;
                         oBrowse:setPos( {oBrowse:currentPos()[1],oStaticQuery:currentSize()[2]}),;
                         oStaticQuery:show(), ;
                         DC_GetRefresh(::queryGetList), ;
                         SetAppFocus(DC_GetObject(::queryGetList,'QUERY_1'))))}

  ::doneBlock := {|| oStaticQuery:hide(), ;
                   oBrowse:setSize( oStaticBrowse:currentSize() ), ;
                   oBrowse:setPos( {0,0}), DC_GetRefresh(::queryGetList) }
ENDIF

(::alias)->(dbGoTo(nSaveRec))

IF ::isFloating
  DCGETOPTIONS COLOR ::bgColor HIDE
  DCREAD GUI FIT TITLE 'Browse Query' EXIT ;
      EVAL {|o|::dialogWindow := o, ;
               ::browseObject := oBrowse} ;
      OPTIONS GetOptions SAVE
ENDIF

RETURN nil

* ------------

STATIC FUNCTION _GetEvalBlock( aQueryItem, aGetListItem )

RETURN {|o|aQueryItem[QUERY_VAR] := aGetListItem[bGETLIST_VAR]}

* ------------

METHOD DC_BrowseQuery:isVisible()

IF ::isFloating
  RETURN ::dialogWindow:isVisible()
ENDIF

RETURN ::parentObject:isVisible()

* ------------

METHOD DC_BrowseQuery:destroy()

IF Valtype(::dialogWindow) == 'O'
  ::dialogWindow:destroy()
ENDIF

RETURN self

* ------------

METHOD DC_BrowseQuery:applyQuery( oBrowse )

LOCAL i, oDlg, aRecords[0], lAdd, nCnt, nPct, nLastRec, oProgress, GetList[0], GetOptions, ;
      xScope, lCancelled := .f.

IF !::testQuery()
  DC_WinAlert('No Search Criteria Entered!!')
  RETURN nil
ENDIF

nLastRec := (::alias)->(LastRec())
nPct := Round(nLastRec/100,0)
nCnt := 0

@ 0,0 DCPROGRESS oProgress RADIUS 20 DYNAMIC OUTLINE COLOR GRA_CLR_CYAN, GRA_CLR_WHITE SIZE 50,1 PERCENT
@ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Cancel' SIZE 10,1 CONFIG ::buttonConfig ;
   ACTION {||lCancelled := .t.}

DCGETOPTIONS COLOR GRA_CLR_WHITE
DCREAD GUI FIT TITLE 'Searching...' EXIT PARENT @oDlg OPTIONS GetOptions

DC_PointerWait()
(::alias)->(DC_ClrScope())
(::alias)->(DC_SetScopeArray(nil))

FOR i := 1 TO Len(::queryArray)
  IF !Empty(::queryArray[i,QUERY_GET]) .AND. Upper(DC_XtoC(::queryArray[i,QUERY_INDEX])) == (::alias)->(OrdSetFocus())
    IF Valtype(::queryArray[i,QUERY_GET]) = 'C'
      xScope := Trim(::queryArray[i,QUERY_GET])
    ELSEIF Valtype(::queryArray[i,QUERY_GET]) = 'D'
      xScope := Dtos(::queryArray[i,QUERY_GET])
    ELSE
      xScope := ::queryArray[i,QUERY_GET]
    ENDIF
    (::alias)->(DC_SetScope(0,xScope))
    (::alias)->(DC_SetScope(1,xScope))
    EXIT
  ENDIF
NEXT

(::alias)->(DC_dbGoTop())

DO WHILE !(::alias)->(Eof()) .AND. !lCancelled
  IF nCnt % 100 == 0
    DC_CompleteEvents()
  ENDIF
  IF ++nCnt%nPct == 0
    DC_GetProgress( oProgress, nCnt, nLastRec )
  ENDIF
  lAdd := .f.
  FOR i := 1 TO Len(::queryArray)
    IF !Empty(::queryArray[i,QUERY_GET])
      IF Valtype(::queryArray[i,QUERY_GET]) = 'C' .AND. Upper((::alias)->(&(::queryArray[i,QUERY_FIELD]))) = Upper(Trim(::queryArray[i,QUERY_GET]))
        lAdd := .t.
      ELSEIF Valtype(::queryArray[i,QUERY_GET]) = 'D' .AND. DTOS((::alias)->(&(::queryArray[i,QUERY_FIELD]))) = DTOS(::queryArray[i,QUERY_GET])
        lAdd := .t.
      ELSEIF (::alias)->(&(::queryArray[i,QUERY_FIELD])) == ::queryArray[i,QUERY_GET]
        lAdd := .t.
      ELSE
        lAdd := .f.
        EXIT
      ENDIF
    ENDIF
  NEXT
  IF lAdd
    AAdd(aRecords,(::alias)->(RecNo()))
  ENDIF
  (::alias)->(DC_DbSkip())
ENDDO
oDlg:destroy()

IF Empty(aRecords)
  aRecords := {-1 }
ENDIF

(::alias)->(DC_ClrScope())
IF lCancelled
  (::alias)->(DC_SetScopeArray(nil))
ELSE
  (::alias)->(DC_SetScopeArray(aRecords))
ENDIF
(::alias)->(DC_DbGoTop())

oBrowse:refreshAll()
DC_PointerArrow()

IF aRecords[1] = -1 .AND. !lCancelled
  DCMSGBOX "No Matching Records Were Found" TITLE "Data Alert" TIMEOUT 3
ENDIF

RETURN nil

* -----------

METHOD DC_BrowseQuery:testQuery()

LOCAL i, lStatus := .f.

FOR i := 1 TO Len(::queryArray)
  IF !Empty(::queryArray[i,QUERY_GET])
    lStatus := .t.
    EXIT
  ENDIF
NEXT

RETURN lStatus

* -----------

METHOD DC_BrowseQuery:clearFields( oBrowse )

LOCAL i

(::alias)->(dbGoTo(0))

FOR i := 1 TO Len(::queryArray)
  ::queryArray[i,QUERY_GET] := (::alias)->(&(::queryArray[i,QUERY_FIELD]))
NEXT

(::alias)->(DC_SetScopeArray(nil))
(::alias)->(DC_DbGoTop())

DC_GetRefresh(::queryGetList)
DC_GetRefresh(::mainGetList)
// oBrowse:refreshAll()

RETURN nil

* ------------

GETSETFUNCTION DC_BrowseTrapErrors DEFAULT .f.

* ------------

FUNCTION DC_BrowseConfigMessageBlock( bMsg )

STATIC sbMsg
LOCAL bMsgSave

bMsgSave := sbMsg

IF Valtype(bMsg) == 'B'
  sbMsg := bMsg
ENDIF

RETURN bMsgSave
